
import pkg_resources

from pathspider.cmd.measure import job_feeder_csv as job_feeder

class FakeSpider:
    def __init__(self):
        self.was_shutdown = False
        self.jobs = []

    def add_job(self, row):
        self.jobs.append(row)

    def shutdown(self):
        self.was_shutdown = True

def test_job_feeder_csv_webtest():
    spider = FakeSpider()

    expected_jobs = [{'dip': '160.85.31.173', 'dp': '80', 'domain': 'mami-project.eu', 'rank': '1'},
                     {'dip': '139.133.210.32', 'dp': '80', 'domain': 'erg.abdn.ac.uk', 'rank': '2'},
                     {'dip': '2001:630:241:210:569f:35ff:fe0a:116a', 'dp': '80', 'domain': 'erg.abdn.ac.uk',  'rank': '3'},
                     {'dip': '129.132.52.158', 'dp': '80', 'domain': 'ecn.ethz.ch', 'rank': '4'},
                     {'dip': '2001:67c:10ec:36c2::61', 'dp': '80', 'domain': 'ecn.ethz.ch', 'rank': '5'},
                     {'dip': '139.133.1.4', 'dp': '80', 'domain': 'abdn.ac.uk', 'rank': '6'}]

    job_feeder(pkg_resources.resource_filename("pathspider", "tests/data/webtest.csv"), spider)
    assert spider.was_shutdown
    print (spider.jobs)
    print (expected_jobs)
    assert spider.jobs == expected_jobs

def test_job_feeder_csv_webtest_newline():
    spider = FakeSpider()

    expected_jobs = [{'dip': '160.85.31.173', 'dp': '80', 'domain': 'mami-project.eu', 'rank': '1'},
                     {'dip': '139.133.210.32', 'dp': '80', 'domain': 'erg.abdn.ac.uk', 'rank': '2'},
                     {'dip': '2001:630:241:210:569f:35ff:fe0a:116a', 'dp': '80', 'domain': 'erg.abdn.ac.uk',  'rank': '3'},
                     {'dip': '129.132.52.158', 'dp': '80', 'domain': 'ecn.ethz.ch', 'rank': '4'},
                     {'dip': '2001:67c:10ec:36c2::61', 'dp': '80', 'domain': 'ecn.ethz.ch', 'rank': '5'},
                     {'dip': '139.133.1.4', 'dp': '80', 'domain': 'abdn.ac.uk', 'rank': '6'}]

    job_feeder(pkg_resources.resource_filename("pathspider", "tests/data/webtest_newline.csv"), spider)
    assert spider.was_shutdown
    assert spider.jobs == expected_jobs

def test_job_feeder_csv_webtest_duplicates():
    spider = FakeSpider()

    expected_jobs = [{'dip': '160.85.31.173', 'dp': '80', 'domain': 'mami-project.eu', 'rank': '1'},
                     {'dip': '139.133.210.32', 'dp': '80', 'domain': 'erg.abdn.ac.uk', 'rank': '2'},
                     {'dip': '2001:630:241:210:569f:35ff:fe0a:116a', 'dp': '80', 'domain': 'erg.abdn.ac.uk',  'rank': '3'},
                     {'dip': '129.132.52.158', 'dp': '80', 'domain': 'ecn.ethz.ch', 'rank': '4'},
                     {'dip': '2001:67c:10ec:36c2::61', 'dp': '80', 'domain': 'ecn.ethz.ch', 'rank': '5'},
                     {'dip': '139.133.1.4', 'dp': '80', 'domain': 'abdn.ac.uk', 'rank': '6'}]

    job_feeder(pkg_resources.resource_filename("pathspider", "tests/data/webtest_duplicates.csv"), spider)
    assert spider.was_shutdown
    print (spider.jobs)
    print (expected_jobs)
    assert spider.jobs == expected_jobs

def test_job_feeder_csv_webtest_09x():
    spider = FakeSpider()

    expected_jobs = [{'dip': '160.85.31.173', 'dp': '80', 'domain': 'mami-project.eu'},
                     {'dip': '139.133.210.32', 'dp': '80', 'domain': 'erg.abdn.ac.uk'},
                     {'dip': '2001:630:241:210:569f:35ff:fe0a:116a', 'dp': '80', 'domain': 'erg.abdn.ac.uk'},
                     {'dip': '129.132.52.158', 'dp': '80', 'domain': 'ecn.ethz.ch'},
                     {'dip': '2001:67c:10ec:36c2::61', 'dp': '80', 'domain': 'ecn.ethz.ch'},
                     {'dip': '139.133.1.4', 'dp': '80', 'domain': 'abdn.ac.uk'}]

    job_feeder(pkg_resources.resource_filename("pathspider", "tests/data/webtest_0.9.x.csv"), spider)
    assert spider.was_shutdown
    print (spider.jobs)
    print (expected_jobs)
    assert spider.jobs == expected_jobs

def test_job_feeder_csv_dnsresolvtest():
    spider = FakeSpider()

    expected_jobs = [{'rank': '1', 'domain': 'google.com'},
                     {'rank': '2', 'domain': 'youtube.com'},
                     {'rank': '3', 'domain': 'facebook.com'},
                     {'rank': '4', 'domain': 'baidu.com'},
                     {'rank': '5', 'domain': 'wikipedia.org'},
                     {'rank': '6', 'domain': 'yahoo.com'},
                     {'rank': '7', 'domain': 'google.co.in'},
                     {'rank': '8', 'domain': 'twitter.com'},
                     {'rank': '9', 'domain': 'amazon.com'},
                     {'rank': '10', 'domain': 'qq.com'},
                     {'rank': '11', 'domain': 'google.co.jp'},
                     {'rank': '12', 'domain': 'live.com'},
                     {'rank': '13', 'domain': 'linkedin.com'},
                     {'rank': '14', 'domain': 'taobao.com'},
                     {'rank': '15', 'domain': 'instagram.com'},
                     {'rank': '16', 'domain': 'vk.com'},
                     {'rank': '17', 'domain': 'hao123.com'},
                     {'rank': '18', 'domain': 'sohu.com'},
                     {'rank': '19', 'domain': 'sina.com.cn'},
                     {'rank': '20', 'domain': 'weibo.com'},
                     {'rank': '21', 'domain': '360.cn'},
                     {'rank': '22', 'domain': 'google.de'},
                     {'rank': '23', 'domain': 'reddit.com'},
                     {'rank': '24', 'domain': 'yahoo.co.jp'},
                     {'rank': '25', 'domain': 'google.com.br'},
                     {'rank': '26', 'domain': 'tmall.com'},
                     {'rank': '27', 'domain': 'google.co.uk'},
                     {'rank': '28', 'domain': 'yandex.ru'},
                     {'rank': '29', 'domain': 'google.fr'},
                     {'rank': '30', 'domain': 't.co'},
                     {'rank': '31', 'domain': 'onclickads.net'},
                     {'rank': '32', 'domain': 'google.ru'},
                     {'rank': '33', 'domain': 'ebay.com'},
                     {'rank': '34', 'domain': 'wordpress.com'},
                     {'rank': '35', 'domain': 'soso.com'},
                     {'rank': '36', 'domain': 'stackoverflow.com'},
                     {'rank': '37', 'domain': 'bing.com'},
                     {'rank': '38', 'domain': 'tumblr.com'},
                     {'rank': '39', 'domain': 'blogspot.com'},
                     {'rank': '40', 'domain': 'google.it'},
                     {'rank': '41', 'domain': 'google.es'},
                     {'rank': '42', 'domain': 'msn.com'},
                     {'rank': '43', 'domain': 'microsoft.com'},
                     {'rank': '44', 'domain': 'apple.com'},
                     {'rank': '45', 'domain': 'pinterest.com'},
                     {'rank': '46', 'domain': 'aliexpress.com'},
                     {'rank': '47', 'domain': 'paypal.com'},
                     {'rank': '48', 'domain': 'netflix.com'},
                     {'rank': '49', 'domain': 'google.com.mx'},
                     {'rank': '50', 'domain': 'blogger.com'},
                     {'rank': '51', 'domain': 'gmw.cn'},
                     {'rank': '52', 'domain': 'github.com'},
                     {'rank': '53', 'domain': 'imgur.com'},
                     {'rank': '54', 'domain': 'imdb.com'},
                     {'rank': '55', 'domain': 'mail.ru'},
                     {'rank': '56', 'domain': 'amazon.co.jp'},
                     {'rank': '57', 'domain': 'google.ca'},
                     {'rank': '58', 'domain': 'whatsapp.com'},
                     {'rank': '59', 'domain': 'naver.com'},
                     {'rank': '60', 'domain': 'google.com.hk'},
                     {'rank': '61', 'domain': 'amazon.in'},
                     {'rank': '62', 'domain': 'popads.net'},
                     {'rank': '63', 'domain': 'google.co.kr'},
                     {'rank': '64', 'domain': 'pornhub.com'},
                     {'rank': '65', 'domain': 'ok.ru'},
                     {'rank': '66', 'domain': 'office.com'},
                     {'rank': '67', 'domain': 'cnzz.com'},
                     {'rank': '68', 'domain': 'google.co.id'},
                     {'rank': '69', 'domain': 'fc2.com'},
                     {'rank': '70', 'domain': 'dropbox.com'},
                     {'rank': '71', 'domain': 'youku.com'},
                     {'rank': '72', 'domain': 'xvideos.com'},
                     {'rank': '73', 'domain': 'google.com.tr'},
                     {'rank': '74', 'domain': 'googleusercontent.com'},
                     {'rank': '75', 'domain': 'microsoftonline.com'},
                     {'rank': '76', 'domain': 'google.com.au'},
                     {'rank': '77', 'domain': 'google.pl'},
                     {'rank': '78', 'domain': 'google.com.tw'},
                     {'rank': '79', 'domain': 'xhamster.com'},
                     {'rank': '80', 'domain': 'quora.com'},
                     {'rank': '81', 'domain': 'cnn.com'},
                     {'rank': '82', 'domain': 'adobe.com'},
                     {'rank': '83', 'domain': 'xinhuanet.com'},
                     {'rank': '84', 'domain': 'flipkart.com'},
                     {'rank': '85', 'domain': 'tianya.cn'},
                     {'rank': '86', 'domain': 'amazon.de'},
                     {'rank': '87', 'domain': 'bongacams.com'},
                     {'rank': '88', 'domain': 'twitch.tv'},
                     {'rank': '89', 'domain': 'youth.cn'},
                     {'rank': '90', 'domain': 'craigslist.org'},
                     {'rank': '91', 'domain': 'pixnet.net'},
                     {'rank': '92', 'domain': 'wikia.com'},
                     {'rank': '93', 'domain': 'alibaba.com'},
                     {'rank': '94', 'domain': 'godaddy.com'},
                     {'rank': '95', 'domain': '360.com'},
                     {'rank': '96', 'domain': 'isanalyze.com'},
                     {'rank': '97', 'domain': 'google.com.eg'},
                     {'rank': '98', 'domain': 'coccoc.com'},
                     {'rank': '99', 'domain': 'google.com.ar'},
                     {'rank': '100', 'domain': 'haosou.com'},
                     {'rank': '101', 'domain': 'adf.ly'},
                     {'rank': '102', 'domain': 'bbc.com'},
                     {'rank': '103', 'domain': 'so.com'},
                     {'rank': '104', 'domain': 'amazon.co.uk'},
                     {'rank': '105', 'domain': 'rakuten.co.jp'},
                     {'rank': '106', 'domain': 'google.com.sa'},
                     {'rank': '107', 'domain': 'salesforce.com'},
                     {'rank': '108', 'domain': 'amazonaws.com'},
                     {'rank': '109', 'domain': 'google.com.pk'},
                     {'rank': '110', 'domain': 'google.co.th'},
                     {'rank': '111', 'domain': 'terraclicks.com'},
                     {'rank': '112', 'domain': 'vimeo.com'},
                     {'rank': '113', 'domain': 'espn.com'},
                     {'rank': '114', 'domain': 'nytimes.com'},
                     {'rank': '115', 'domain': 'bbc.co.uk'},
                     {'rank': '116', 'domain': 'booking.com'},
                     {'rank': '117', 'domain': 'soundcloud.com'},
                     {'rank': '118', 'domain': 'indiatimes.com'},
                     {'rank': '119', 'domain': 'ask.com'},
                     {'rank': '120', 'domain': 'ebay.de'},
                     {'rank': '121', 'domain': 'txxx.com'},
                     {'rank': '122', 'domain': 'google.nl'},
                     {'rank': '123', 'domain': 'people.com.cn'},
                     {'rank': '124', 'domain': 'dailymotion.com'},
                     {'rank': '125', 'domain': 'globo.com'},
                     {'rank': '126', 'domain': 'uol.com.br'},
                     {'rank': '127', 'domain': 'w3schools.com'},
                     {'rank': '128', 'domain': 'china.com.cn'},
                     {'rank': '129', 'domain': 'ebay.co.uk'},
                     {'rank': '130', 'domain': 'stackexchange.com'},
                     {'rank': '131', 'domain': 'alipay.com'},
                     {'rank': '132', 'domain': 'diply.com'},
                     {'rank': '133', 'domain': 'zhihu.com'},
                     {'rank': '134', 'domain': 'dailymail.co.uk'},
                     {'rank': '135', 'domain': 'daum.net'},
                     {'rank': '136', 'domain': 'tudou.com'},
                     {'rank': '137', 'domain': 'thepiratebay.org'},
                     {'rank': '138', 'domain': 'bet365.com'},
                     {'rank': '139', 'domain': 'wordpress.org'},
                     {'rank': '140', 'domain': 'outbrain.com'},
                     {'rank': '141', 'domain': 'ettoday.net'},
                     {'rank': '142', 'domain': 'google.co.ve'},
                     {'rank': '143', 'domain': 'savefrom.net'},
                     {'rank': '144', 'domain': 'indeed.com'},
                     {'rank': '145', 'domain': 'nicovideo.jp'},
                     {'rank': '146', 'domain': 'force.com'},
                     {'rank': '147', 'domain': 'google.co.za'},
                     {'rank': '148', 'domain': 'trello.com'},
                     {'rank': '149', 'domain': 'deviantart.com'},
                     {'rank': '150', 'domain': 'theguardian.com'},
                     {'rank': '151', 'domain': '9gag.com'},
                     {'rank': '152', 'domain': 'slideshare.net'},
                     {'rank': '153', 'domain': 'chase.com'},
                     {'rank': '154', 'domain': 'huffingtonpost.com'},
                     {'rank': '155', 'domain': 'blastingnews.com'},
                     {'rank': '156', 'domain': 'buzzfeed.com'},
                     {'rank': '157', 'domain': 'uptodown.com'},
                     {'rank': '158', 'domain': 'vice.com'},
                     {'rank': '159', 'domain': 'doubleclick.net'},
                     {'rank': '160', 'domain': 'slack.com'},
                     {'rank': '161', 'domain': 'bp.blogspot.com'},
                     {'rank': '162', 'domain': 'github.io'},
                     {'rank': '163', 'domain': 'avito.ru'},
                     {'rank': '164', 'domain': 'google.com.ph'},
                     {'rank': '165', 'domain': 'livejasmin.com'},
                     {'rank': '166', 'domain': 'cnet.com'},
                     {'rank': '167', 'domain': 'thewhizmarketing.com'},
                     {'rank': '168', 'domain': 'bilibili.com'},
                     {'rank': '169', 'domain': 'tribunnews.com'},
                     {'rank': '170', 'domain': 'google.com.co'},
                     {'rank': '171', 'domain': 'steampowered.com'},
                     {'rank': '172', 'domain': 'snapdeal.com'},
                     {'rank': '173', 'domain': 'livedoor.jp'},
                     {'rank': '174', 'domain': 'mediafire.com'},
                     {'rank': '175', 'domain': 'google.com.vn'},
                     {'rank': '176', 'domain': 'etsy.com'},
                     {'rank': '177', 'domain': 'china.com'},
                     {'rank': '178', 'domain': 'blogspot.in'},
                     {'rank': '179', 'domain': 'popcash.net'},
                     {'rank': '180', 'domain': 'google.gr'},
                     {'rank': '181', 'domain': 'ameblo.jp'},
                     {'rank': '182', 'domain': 'sogou.com'},
                     {'rank': '183', 'domain': 'washingtonpost.com'},
                     {'rank': '184', 'domain': 'huanqiu.com'},
                     {'rank': '185', 'domain': 'shutterstock.com'},
                     {'rank': '186', 'domain': 'addthis.com'},
                     {'rank': '187', 'domain': 'chaturbate.com'},
                     {'rank': '188', 'domain': 'google.com.ng'},
                     {'rank': '189', 'domain': 'fbcdn.net'},
                     {'rank': '190', 'domain': 'onlinesbi.com'},
                     {'rank': '191', 'domain': 'spotify.com'},
                     {'rank': '192', 'domain': 'extratorrent.cc'},
                     {'rank': '193', 'domain': 'mozilla.org'},
                     {'rank': '194', 'domain': 'skype.com'},
                     {'rank': '195', 'domain': 'weather.com'},
                     {'rank': '196', 'domain': 'bankofamerica.com'},
                     {'rank': '197', 'domain': 'hclips.com'},
                     {'rank': '198', 'domain': 'naver.jp'},
                     {'rank': '199', 'domain': 'google.com.sg'},
                     {'rank': '200', 'domain': 'google.be'},
                     {'rank': '201', 'domain': 'twimg.com'},
                     {'rank': '202', 'domain': 'google.ro'},
                     {'rank': '203', 'domain': 'wikihow.com'},
                     {'rank': '204', 'domain': 'cloudfront.net'},
                     {'rank': '205', 'domain': 'google.com.ua'},
                     {'rank': '206', 'domain': 'forbes.com'},
                     {'rank': '207', 'domain': 'mercadolivre.com.br'},
                     {'rank': '208', 'domain': 'goo.ne.jp'},
                     {'rank': '209', 'domain': 'blogspot.co.id'},
                     {'rank': '210', 'domain': 'cctv.com'},
                     {'rank': '211', 'domain': 'babytree.com'},
                     {'rank': '212', 'domain': 'livejournal.com'},
                     {'rank': '213', 'domain': 'csdn.net'},
                     {'rank': '214', 'domain': 'adexchangeprediction.com'},
                     {'rank': '215', 'domain': 'ozock.com'},
                     {'rank': '216', 'domain': 'tistory.com'},
                     {'rank': '217', 'domain': 'detik.com'},
                     {'rank': '218', 'domain': 'xnxx.com'},
                     {'rank': '219', 'domain': 'google.se'},
                     {'rank': '220', 'domain': 'wetransfer.com'},
                     {'rank': '221', 'domain': 'amazon.it'},
                     {'rank': '222', 'domain': 'walmart.com'},
                     {'rank': '223', 'domain': 'wordreference.com'},
                     {'rank': '224', 'domain': 'giphy.com'},
                     {'rank': '225', 'domain': 'wellsfargo.com'},
                     {'rank': '226', 'domain': 'zol.com.cn'},
                     {'rank': '227', 'domain': 'archive.org'},
                     {'rank': '228', 'domain': 'feedly.com'},
                     {'rank': '229', 'domain': 'redtube.com'},
                     {'rank': '230', 'domain': 'amazon.fr'},
                     {'rank': '231', 'domain': 'wikimedia.org'},
                     {'rank': '232', 'domain': '51.la'},
                     {'rank': '233', 'domain': 'flickr.com'},
                     {'rank': '234', 'domain': 'foxnews.com'},
                     {'rank': '235', 'domain': 'nih.gov'},
                     {'rank': '236', 'domain': 'google.com.pe'},
                     {'rank': '237', 'domain': 'openload.co'},
                     {'rank': '238', 'domain': 'gfycat.com'},
                     {'rank': '239', 'domain': 'upornia.com'},
                     {'rank': '240', 'domain': 'google.dz'},
                     {'rank': '241', 'domain': 'yelp.com'},
                     {'rank': '242', 'domain': 'files.wordpress.com'},
                     {'rank': '243', 'domain': 'google.at'},
                     {'rank': '244', 'domain': 'kinogo.club'},
                     {'rank': '245', 'domain': '163.com'},
                     {'rank': '246', 'domain': 'weebly.com'},
                     {'rank': '247', 'domain': 'mama.cn'},
                     {'rank': '248', 'domain': 'themeforest.net'},
                     {'rank': '249', 'domain': 'pinimg.com'},
                     {'rank': '250', 'domain': 'ikea.com'},
                     {'rank': '251', 'domain': 'paytm.com'},
                     {'rank': '252', 'domain': 'mailchimp.com'},
                     {'rank': '253', 'domain': 'tripadvisor.com'},
                     {'rank': '254', 'domain': 'aol.com'},
                     {'rank': '255', 'domain': 'taboola.com'},
                     {'rank': '256', 'domain': 'freepik.com'},
                     {'rank': '257', 'domain': 'reimageplus.com'},
                     {'rank': '258', 'domain': 'putlocker.is'},
                     {'rank': '259', 'domain': 'rutracker.org'},
                     {'rank': '260', 'domain': 'instructure.com'},
                     {'rank': '261', 'domain': 'tuberel.com'},
                     {'rank': '262', 'domain': 'detail.tmall.com'},
                     {'rank': '263', 'domain': 'allegro.pl'},
                     {'rank': '264', 'domain': 'sourceforge.net'},
                     {'rank': '265', 'domain': 'businessinsider.com'},
                     {'rank': '266', 'domain': 'zendesk.com'},
                     {'rank': '267', 'domain': 'google.ch'},
                     {'rank': '268', 'domain': 'theladbible.com'},
                     {'rank': '269', 'domain': 'iwanttodeliver.com'},
                     {'rank': '270', 'domain': 'ndtv.com'},
                     {'rank': '271', 'domain': 'google.pt'},
                     {'rank': '272', 'domain': 'myway.com'},
                     {'rank': '273', 'domain': 'youporn.com'},
                     {'rank': '274', 'domain': 'espncricinfo.com'},
                     {'rank': '275', 'domain': 'dmm.co.jp'},
                     {'rank': '276', 'domain': 'google.ae'},
                     {'rank': '277', 'domain': 'douban.com'},
                     {'rank': '278', 'domain': 'upwork.com'},
                     {'rank': '279', 'domain': 'hdfcbank.com'},
                     {'rank': '280', 'domain': 'zillow.com'},
                     {'rank': '281', 'domain': 'varzesh3.com'},
                     {'rank': '282', 'domain': 'google.az'},
                     {'rank': '283', 'domain': 'xywy.com'},
                     {'rank': '284', 'domain': 'google.cz'},
                     {'rank': '285', 'domain': 'google.cn'},
                     {'rank': '286', 'domain': 'thesaurus.com'},
                     {'rank': '287', 'domain': 'speedtest.net'},
                     {'rank': '288', 'domain': 'youm7.com'},
                     {'rank': '289', 'domain': 'zippyshare.com'},
                     {'rank': '290', 'domain': 'rt.com'},
                     {'rank': '291', 'domain': 'doublepimp.com'},
                     {'rank': '292', 'domain': 'webtretho.com'},
                     {'rank': '293', 'domain': 'medium.com'},
                     {'rank': '294', 'domain': 'google.com.bd'},
                     {'rank': '295', 'domain': 'disqus.com'},
                     {'rank': '296', 'domain': 'nametests.com'},
                     {'rank': '297', 'domain': 'globaloffers.link'},
                     {'rank': '298', 'domain': 'scribd.com'},
                     {'rank': '299', 'domain': 'fiverr.com'},
                     {'rank': '300', 'domain': 'bitauto.com'},
                     {'rank': '301', 'domain': 'douyu.com'},
                     {'rank': '302', 'domain': 'kinopoisk.ru'},
                     {'rank': '303', 'domain': 'gsmarena.com'},
                     {'rank': '304', 'domain': 'softonic.com'},
                     {'rank': '305', 'domain': 'google.cl'},
                     {'rank': '306', 'domain': 'sh.st'},
                     {'rank': '307', 'domain': 'icicibank.com'},
                     {'rank': '308', 'domain': 'mega.nz'},
                     {'rank': '309', 'domain': 'gearbest.com'},
                     {'rank': '310', 'domain': 'zoho.com'},
                     {'rank': '311', 'domain': 'secureserver.net'},
                     {'rank': '312', 'domain': 'media.tumblr.com'},
                     {'rank': '313', 'domain': 'nfl.com'},
                     {'rank': '314', 'domain': 'caijing.com.cn'},
                     {'rank': '315', 'domain': 'rambler.ru'},
                     {'rank': '316', 'domain': 'leboncoin.fr'},
                     {'rank': '317', 'domain': 'kakaku.com'},
                     {'rank': '318', 'domain': 'amazon.es'},
                     {'rank': '319', 'domain': 'daikynguyenvn.com'},
                     {'rank': '320', 'domain': 'rarbg.to'},
                     {'rank': '321', 'domain': 'ebay-kleinanzeigen.de'},
                     {'rank': '322', 'domain': 'oracle.com'},
                     {'rank': '323', 'domain': 'wix.com'},
                     {'rank': '324', 'domain': '39.net'},
                     {'rank': '325', 'domain': 'hatenablog.com'},
                     {'rank': '326', 'domain': 'spotscenered.info'},
                     {'rank': '327', 'domain': 'behance.net'},
                     {'rank': '328', 'domain': 'messenger.com'},
                     {'rank': '329', 'domain': 'directrev.com'},
                     {'rank': '330', 'domain': 'evernote.com'},
                     {'rank': '331', 'domain': 'asos.com'},
                     {'rank': '332', 'domain': 'iqiyi.com'},
                     {'rank': '333', 'domain': 'onclickpredictiv.com'},
                     {'rank': '334', 'domain': 'tokopedia.com'},
                     {'rank': '335', 'domain': 'jd.com'},
                     {'rank': '336', 'domain': 'google.no'},
                     {'rank': '337', 'domain': 'icloud.com'},
                     {'rank': '338', 'domain': 'hatena.ne.jp'},
                     {'rank': '339', 'domain': 'goodreads.com'},
                     {'rank': '340', 'domain': 'ytimg.com'},
                     {'rank': '341', 'domain': '51yes.com'},
                     {'rank': '342', 'domain': 'bloomberg.com'},
                     {'rank': '343', 'domain': '1905.com'},
                     {'rank': '344', 'domain': 'hootsuite.com'},
                     {'rank': '345', 'domain': 'go.com'},
                     {'rank': '346', 'domain': 'hp.com'},
                     {'rank': '347', 'domain': 'canva.com'},
                     {'rank': '348', 'domain': 'chinadaily.com.cn'},
                     {'rank': '349', 'domain': 'cnblogs.com'},
                     {'rank': '350', 'domain': 'samsung.com'},
                     {'rank': '351', 'domain': 'adnetworkperformance.com'},
                     {'rank': '352', 'domain': 'shopify.com'},
                     {'rank': '353', 'domain': 'repubblica.it'},
                     {'rank': '354', 'domain': 'battle.net'},
                     {'rank': '355', 'domain': 'onedio.com'},
                     {'rank': '356', 'domain': 'tradeadexchange.com'},
                     {'rank': '357', 'domain': '1688.com'},
                     {'rank': '358', 'domain': 'google.hu'},
                     {'rank': '359', 'domain': 'youtube-mp3.org'},
                     {'rank': '360', 'domain': 'atlassian.net'},
                     {'rank': '361', 'domain': 'ltn.com.tw'},
                     {'rank': '362', 'domain': 'sharepoint.com'},
                     {'rank': '363', 'domain': '123movies.to'},
                     {'rank': '364', 'domain': 'telegraph.co.uk'},
                     {'rank': '365', 'domain': 'google.ie'},
                     {'rank': '366', 'domain': 'yesky.com'},
                     {'rank': '367', 'domain': 'cnnic.cn'},
                     {'rank': '368', 'domain': 'americanexpress.com'},
                     {'rank': '369', 'domain': 'kompas.com'},
                     {'rank': '370', 'domain': 'onet.pl'},
                     {'rank': '371', 'domain': '2ch.net'},
                     {'rank': '372', 'domain': 'googleapis.com'},
                     {'rank': '373', 'domain': 'amazon.cn'},
                     {'rank': '374', 'domain': 'pandora.com'},
                     {'rank': '375', 'domain': 'steamcommunity.com'},
                     {'rank': '376', 'domain': 'eksisozluk.com'},
                     {'rank': '377', 'domain': 'nyaa.se'},
                     {'rank': '378', 'domain': 'roblox.com'},
                     {'rank': '379', 'domain': 'quizlet.com'},
                     {'rank': '380', 'domain': 'kaskus.co.id'},
                     {'rank': '381', 'domain': 'blkget.com'},
                     {'rank': '382', 'domain': 'gmx.net'},
                     {'rank': '383', 'domain': 'bukalapak.com'},
                     {'rank': '384', 'domain': 'xfinity.com'},
                     {'rank': '385', 'domain': 'google.co.il'},
                     {'rank': '386', 'domain': 'enet.com.cn'},
                     {'rank': '387', 'domain': 'elpais.com'},
                     {'rank': '388', 'domain': 'pixabay.com'},
                     {'rank': '389', 'domain': 'irctc.co.in'},
                     {'rank': '390', 'domain': 'blogspot.jp'},
                     {'rank': '391', 'domain': 'bestbuy.com'},
                     {'rank': '392', 'domain': 'wsj.com'},
                     {'rank': '393', 'domain': '4dsply.com'},
                     {'rank': '394', 'domain': 'web.de'},
                     {'rank': '395', 'domain': 'udemy.com'},
                     {'rank': '396', 'domain': 'kissanime.to'},
                     {'rank': '397', 'domain': 'dictionary.com'},
                     {'rank': '398', 'domain': 'usatoday.com'},
                     {'rank': '399', 'domain': 'thewhizproducts.com'},
                     {'rank': '400', 'domain': 'seasonvar.ru'},
                     {'rank': '401', 'domain': 'fastpokemap.se'},
                     {'rank': '402', 'domain': 'telegram.org'},
                     {'rank': '403', 'domain': 'ebay.in'},
                     {'rank': '404', 'domain': 'reddituploads.com'},
                     {'rank': '405', 'domain': 'infusionsoft.com'},
                     {'rank': '406', 'domain': 'sberbank.ru'},
                     {'rank': '407', 'domain': 'liputan6.com'},
                     {'rank': '408', 'domain': 'yts.ag'},
                     {'rank': '409', 'domain': 'accuweather.com'},
                     {'rank': '410', 'domain': 'airbnb.com'},
                     {'rank': '411', 'domain': 'about.com'},
                     {'rank': '412', 'domain': 'blackboard.com'},
                     {'rank': '413', 'domain': 'conservativetribune.com'},
                     {'rank': '414', 'domain': 'google.dk'},
                     {'rank': '415', 'domain': 'mediawhirl.net'},
                     {'rank': '416', 'domain': 'google.fi'},
                     {'rank': '417', 'domain': 'ruten.com.tw'},
                     {'rank': '418', 'domain': 'sabah.com.tr'},
                     {'rank': '419', 'domain': 'weblio.jp'},
                     {'rank': '420', 'domain': 'wittyfeed.com'},
                     {'rank': '421', 'domain': 'tutorialspoint.com'},
                     {'rank': '422', 'domain': 'wixsite.com'},
                     {'rank': '423', 'domain': 'reuters.com'},
                     {'rank': '424', 'domain': 'hubspot.com'},
                     {'rank': '425', 'domain': 'box.com'},
                     {'rank': '426', 'domain': 'clickadu.com'},
                     {'rank': '427', 'domain': 'avg.com'},
                     {'rank': '428', 'domain': 'usps.com'},
                     {'rank': '429', 'domain': 'ups.com'},
                     {'rank': '430', 'domain': 'exoclick.com'},
                     {'rank': '431', 'domain': 'capitalone.com'},
                     {'rank': '432', 'domain': 'techcrunch.com'},
                     {'rank': '433', 'domain': 'sciencedirect.com'},
                     {'rank': '434', 'domain': 'digikala.com'},
                     {'rank': '435', 'domain': 'dell.com'},
                     {'rank': '436', 'domain': 'spiegel.de'},
                     {'rank': '437', 'domain': 'homedepot.com'},
                     {'rank': '438', 'domain': 'php.net'},
                     {'rank': '439', 'domain': 'taleo.net'},
                     {'rank': '440', 'domain': 'kapanlagi.com'},
                     {'rank': '441', 'domain': 'kickstarter.com'},
                     {'rank': '442', 'domain': 'list-manage.com'},
                     {'rank': '443', 'domain': 'ebay.it'},
                     {'rank': '444', 'domain': 'neobux.com'},
                     {'rank': '445', 'domain': 'target.com'},
                     {'rank': '446', 'domain': 'goo.gl'},
                     {'rank': '447', 'domain': 'blog.jp'},
                     {'rank': '448', 'domain': 'indianexpress.com'},
                     {'rank': '449', 'domain': 'adexc.net'},
                     {'rank': '450', 'domain': 'fedex.com'},
                     {'rank': '451', 'domain': 'ign.com'},
                     {'rank': '452', 'domain': 'abs-cbn.com'},
                     {'rank': '453', 'domain': 'thefreedictionary.com'},
                     {'rank': '454', 'domain': 'researchgate.net'},
                     {'rank': '455', 'domain': 'subscene.com'},
                     {'rank': '456', 'domain': 'imzog.com'},
                     {'rank': '457', 'domain': 'intuit.com'},
                     {'rank': '458', 'domain': 'mi.com'},
                     {'rank': '459', 'domain': 'taringa.net'},
                     {'rank': '460', 'domain': 'google.sk'},
                     {'rank': '461', 'domain': 'marca.com'},
                     {'rank': '462', 'domain': 'uploaded.net'},
                     {'rank': '463', 'domain': 'kooora.com'},
                     {'rank': '464', 'domain': 'beeg.com'},
                     {'rank': '465', 'domain': 'bhaskar.com'},
                     {'rank': '466', 'domain': 'baike.com'},
                     {'rank': '467', 'domain': 'asana.com'},
                     {'rank': '468', 'domain': 'wp.pl'},
                     {'rank': '469', 'domain': 'ouo.io'},
                     {'rank': '470', 'domain': 'qiita.com'},
                     {'rank': '471', 'domain': 'mlb.com'},
                     {'rank': '472', 'domain': 'ci123.com'},
                     {'rank': '473', 'domain': 'groupon.com'},
                     {'rank': '474', 'domain': 'spankbang.com'},
                     {'rank': '475', 'domain': 'lifebuzz.com'},
                     {'rank': '476', 'domain': 'mercadolibre.com.ar'},
                     {'rank': '477', 'domain': 'webmd.com'},
                     {'rank': '478', 'domain': 'hotstar.com'},
                     {'rank': '479', 'domain': 'clipconverter.cc'},
                     {'rank': '480', 'domain': 'youdao.com'},
                     {'rank': '481', 'domain': 'orange.fr'},
                     {'rank': '482', 'domain': 'tube8.com'},
                     {'rank': '483', 'domain': 'zone-telechargement.com'},
                     {'rank': '484', 'domain': 'mashable.com'},
                     {'rank': '485', 'domain': 'hurriyet.com.tr'},
                     {'rank': '486', 'domain': 'teepr.com'},
                     {'rank': '487', 'domain': 'eastday.com'},
                     {'rank': '488', 'domain': 'billdesk.com'},
                     {'rank': '489', 'domain': 'free.fr'},
                     {'rank': '490', 'domain': '4chan.org'},
                     {'rank': '491', 'domain': 'azlyrics.com'},
                     {'rank': '492', 'domain': 'ebay.com.au'},
                     {'rank': '493', 'domain': 'bitbucket.org'},
                     {'rank': '494', 'domain': 'bitly.com'},
                     {'rank': '495', 'domain': 'leagueoflegends.com'},
                     {'rank': '496', 'domain': 'amazon.ca'},
                     {'rank': '497', 'domain': '4shared.com'},
                     {'rank': '498', 'domain': 'naukri.com'},
                     {'rank': '499', 'domain': 'zomato.com'},
                     {'rank': '500', 'domain': 'gizmodo.com'},
                     {'rank': '501', 'domain': 'rediff.com'},
                     {'rank': '502', 'domain': 'goal.com'},
                     {'rank': '503', 'domain': 'dcinside.com'},
                     {'rank': '504', 'domain': 'theverge.com'},
                     {'rank': '505', 'domain': 'sahibinden.com'},
                     {'rank': '506', 'domain': 'att.com'},
                     {'rank': '507', 'domain': '17ok.com'},
                     {'rank': '508', 'domain': 'comcast.net'},
                     {'rank': '509', 'domain': 'xe.com'},
                     {'rank': '510', 'domain': 'fivethirtyeight.com'},
                     {'rank': '511', 'domain': 'watsons.tmall.com'},
                     {'rank': '512', 'domain': 'engadget.com'},
                     {'rank': '513', 'domain': 'hola.com'},
                     {'rank': '514', 'domain': 'vk.me'},
                     {'rank': '515', 'domain': 'independent.co.uk'},
                     {'rank': '516', 'domain': 'hulu.com'},
                     {'rank': '517', 'domain': 'elmundo.es'},
                     {'rank': '518', 'domain': 'haber7.com'},
                     {'rank': '519', 'domain': 'citi.com'},
                     {'rank': '520', 'domain': 'nownews.com'},
                     {'rank': '521', 'domain': 'ask.fm'},
                     {'rank': '522', 'domain': 'meetup.com'},
                     {'rank': '523', 'domain': 'gamepedia.com'},
                     {'rank': '524', 'domain': 'mit.edu'},
                     {'rank': '525', 'domain': 'gyazo.com'},
                     {'rank': '526', 'domain': 'hilltopads.net'},
                     {'rank': '527', 'domain': 'corriere.it'},
                     {'rank': '528', 'domain': 'rednet.cn'},
                     {'rank': '529', 'domain': 'tabelog.com'},
                     {'rank': '530', 'domain': 'shopclues.com'},
                     {'rank': '531', 'domain': 'gamefaqs.com'},
                     {'rank': '532', 'domain': 'chatwork.com'},
                     {'rank': '533', 'domain': 'cricbuzz.com'},
                     {'rank': '534', 'domain': 'poptm.com'},
                     {'rank': '535', 'domain': 'pikabu.ru'},
                     {'rank': '536', 'domain': 'blogfa.com'},
                     {'rank': '537', 'domain': 'wowhead.com'},
                     {'rank': '538', 'domain': 'ifeng.com'},
                     {'rank': '539', 'domain': 'xda-developers.com'},
                     {'rank': '540', 'domain': 'intoday.in'},
                     {'rank': '541', 'domain': 'aparat.com'},
                     {'rank': '542', 'domain': 'reverso.net'},
                     {'rank': '543', 'domain': 'livedoor.com'},
                     {'rank': '544', 'domain': 'verizonwireless.com'},
                     {'rank': '545', 'domain': 'aliyun.com'},
                     {'rank': '546', 'domain': 'namu.wiki'},
                     {'rank': '547', 'domain': 'buzzhand.com'},
                     {'rank': '548', 'domain': 't-online.de'},
                     {'rank': '549', 'domain': 'gamer.com.tw'},
                     {'rank': '550', 'domain': 'seznam.cz'},
                     {'rank': '551', 'domain': 'rapidgator.net'},
                     {'rank': '552', 'domain': 'google.com.kw'},
                     {'rank': '553', 'domain': 'chaoshi.tmall.com'},
                     {'rank': '554', 'domain': 'coursera.org'},
                     {'rank': '555', 'domain': 'lemonde.fr'},
                     {'rank': '556', 'domain': 'bild.de'},
                     {'rank': '557', 'domain': 'adidas.tmall.com'},
                     {'rank': '558', 'domain': 'abril.com.br'},
                     {'rank': '559', 'domain': 'milliyet.com.tr'},
                     {'rank': '560', 'domain': 'google.lk'},
                     {'rank': '561', 'domain': 'prezi.com'},
                     {'rank': '562', 'domain': 'hm.com'},
                     {'rank': '563', 'domain': 'google.kz'},
                     {'rank': '564', 'domain': 'surveymonkey.com'},
                     {'rank': '565', 'domain': 'jabong.com'},
                     {'rank': '566', 'domain': 'howtogeek.com'},
                     {'rank': '567', 'domain': 'urdupoint.com'},
                     {'rank': '568', 'domain': 'playstation.com'},
                     {'rank': '569', 'domain': 'wp.com'},
                     {'rank': '570', 'domain': 'souq.com'},
                     {'rank': '571', 'domain': 'clixsense.com'},
                     {'rank': '572', 'domain': 'pantip.com'},
                     {'rank': '573', 'domain': 'line.me'},
                     {'rank': '574', 'domain': 'politico.com'},
                     {'rank': '575', 'domain': 'mercadolibre.com.mx'},
                     {'rank': '576', 'domain': 'gismeteo.ru'},
                     {'rank': '577', 'domain': 'prnt.sc'},
                     {'rank': '578', 'domain': 'airtel.in'},
                     {'rank': '579', 'domain': 'doorblog.jp'},
                     {'rank': '580', 'domain': 'asus.com'},
                     {'rank': '581', 'domain': 'lenta.ru'},
                     {'rank': '582', 'domain': 'umblr.com'},
                     {'rank': '583', 'domain': 'olx.com.br'},
                     {'rank': '584', 'domain': 'sq.cn'},
                     {'rank': '585', 'domain': 'wiktionary.org'},
                     {'rank': '586', 'domain': 'urbandictionary.com'},
                     {'rank': '587', 'domain': 'reference.com'},
                     {'rank': '588', 'domain': 'merdeka.com'},
                     {'rank': '589', 'domain': 'sakura.ne.jp'},
                     {'rank': '590', 'domain': 'superuser.com'},
                     {'rank': '591', 'domain': 'chinaz.com'},
                     {'rank': '592', 'domain': 'drudgereport.com'},
                     {'rank': '593', 'domain': 'nbcnews.com'},
                     {'rank': '594', 'domain': 'state.gov'},
                     {'rank': '595', 'domain': 'uber.com'},
                     {'rank': '596', 'domain': 'instructables.com'},
                     {'rank': '597', 'domain': 'alicdn.com'},
                     {'rank': '598', 'domain': 'lifehacker.com'},
                     {'rank': '599', 'domain': 'wiley.com'},
                     {'rank': '600', 'domain': 'glassdoor.com'},
                     {'rank': '601', 'domain': 'macys.com'},
                     {'rank': '602', 'domain': '51sole.com'},
                     {'rank': '603', 'domain': 'hespress.com'},
                     {'rank': '604', 'domain': 'livescore.com'},
                     {'rank': '605', 'domain': 'nike.com'},
                     {'rank': '606', 'domain': 'digitalocean.com'},
                     {'rank': '607', 'domain': 'seesaa.net'},
                     {'rank': '608', 'domain': 'jrj.com.cn'},
                     {'rank': '609', 'domain': 'google.bg'},
                     {'rank': '610', 'domain': 'ero-advertising.com'},
                     {'rank': '611', 'domain': 'wunderground.com'},
                     {'rank': '612', 'domain': 'ck101.com'},
                     {'rank': '613', 'domain': 'dingit.tv'},
                     {'rank': '614', 'domain': 'bookmyshow.com'},
                     {'rank': '615', 'domain': 'stumbleupon.com'},
                     {'rank': '616', 'domain': 'huaban.com'},
                     {'rank': '617', 'domain': 'freelancer.com'},
                     {'rank': '618', 'domain': 'codepen.io'},
                     {'rank': '619', 'domain': 'cbssports.com'},
                     {'rank': '620', 'domain': 'jimdo.com'},
                     {'rank': '621', 'domain': 'basecamp.com'},
                     {'rank': '622', 'domain': 'expedia.com'},
                     {'rank': '623', 'domain': 'android.com'},
                     {'rank': '624', 'domain': 'blogspot.mx'},
                     {'rank': '625', 'domain': 'issuu.com'},
                     {'rank': '626', 'domain': 'time.com'},
                     {'rank': '627', 'domain': 'yandex.ua'},
                     {'rank': '628', 'domain': 'kdnet.net'},
                     {'rank': '629', 'domain': 'squarespace.com'},
                     {'rank': '630', 'domain': 'investing.com'},
                     {'rank': '631', 'domain': 'mirror.co.uk'},
                     {'rank': '632', 'domain': 'myanimelist.net'},
                     {'rank': '633', 'domain': 'eventbrite.com'},
                     {'rank': '634', 'domain': 'npr.org'},
                     {'rank': '635', 'domain': 'google.co.nz'},
                     {'rank': '636', 'domain': 'photobucket.com'},
                     {'rank': '637', 'domain': 'rottentomatoes.com'},
                     {'rank': '638', 'domain': 'chip.de'},
                     {'rank': '639', 'domain': 'cpasbien.cm'},
                     {'rank': '640', 'domain': 'dmm.com'},
                     {'rank': '641', 'domain': 'springer.com'},
                     {'rank': '642', 'domain': 'filehippo.com'},
                     {'rank': '643', 'domain': 'justdial.com'},
                     {'rank': '644', 'domain': 'viralthread.com'},
                     {'rank': '645', 'domain': 'noaa.gov'},
                     {'rank': '646', 'domain': 'mobile.de'},
                     {'rank': '647', 'domain': 'smartnewtab.com'},
                     {'rank': '648', 'domain': 'spzan.com'},
                     {'rank': '649', 'domain': 'duckduckgo.com'},
                     {'rank': '650', 'domain': 'lenovo.com'},
                     {'rank': '651', 'domain': 'westernjournalism.com'},
                     {'rank': '652', 'domain': 'my-hit.org'},
                     {'rank': '653', 'domain': 'askubuntu.com'},
                     {'rank': '654', 'domain': 'olx.in'},
                     {'rank': '655', 'domain': '4pda.ru'},
                     {'rank': '656', 'domain': 'sitepoint.com'},
                     {'rank': '657', 'domain': 'torrentz2.eu'},
                     {'rank': '658', 'domain': 'curapelanatureza.com.br'},
                     {'rank': '659', 'domain': 'impress.co.jp'},
                     {'rank': '660', 'domain': 'ibm.com'},
                     {'rank': '661', 'domain': 'academia.edu'},
                     {'rank': '662', 'domain': 'drom.ru'},
                     {'rank': '663', 'domain': 'latimes.com'},
                     {'rank': '664', 'domain': 'getpocket.com'},
                     {'rank': '665', 'domain': 'hdzog.com'},
                     {'rank': '666', 'domain': 'japanpost.jp'},
                     {'rank': '667', 'domain': 'commentcamarche.net'},
                     {'rank': '668', 'domain': 'kijiji.ca'},
                     {'rank': '669', 'domain': 'mercadolibre.com.ve'},
                     {'rank': '670', 'domain': 'adbooth.com'},
                     {'rank': '671', 'domain': 'zing.vn'},
                     {'rank': '672', 'domain': 'makeuseof.com'},
                     {'rank': '673', 'domain': 'livedoor.biz'},
                     {'rank': '674', 'domain': 'lazada.co.id'},
                     {'rank': '675', 'domain': 'nikkeibp.co.jp'},
                     {'rank': '676', 'domain': 'e-hentai.org'},
                     {'rank': '677', 'domain': 'adplxmd.com'},
                     {'rank': '678', 'domain': 'myntra.com'},
                     {'rank': '679', 'domain': 'xing.com'},
                     {'rank': '680', 'domain': 'moneycontrol.com'},
                     {'rank': '681', 'domain': 'acfun.tv'},
                     {'rank': '682', 'domain': 'webex.com'},
                     {'rank': '683', 'domain': 'inquirer.net'},
                     {'rank': '684', 'domain': 'newegg.com'},
                     {'rank': '685', 'domain': 'hotels.com'},
                     {'rank': '686', 'domain': 'akamaihd.net'},
                     {'rank': '687', 'domain': 'shink.in'},
                     {'rank': '688', 'domain': 'breitbart.com'},
                     {'rank': '689', 'domain': 'ab4hr.com'},
                     {'rank': '690', 'domain': 'chinaso.com'},
                     {'rank': '691', 'domain': 'answers.com'},
                     {'rank': '692', 'domain': 'lefigaro.fr'},
                     {'rank': '693', 'domain': 'office365.com'},
                     {'rank': '694', 'domain': 'cisco.com'},
                     {'rank': '695', 'domain': '123rf.com'},
                     {'rank': '696', 'domain': 'okezone.com'},
                     {'rank': '697', 'domain': 'banggood.com'},
                     {'rank': '698', 'domain': 'dafont.com'},
                     {'rank': '699', 'domain': 'timeanddate.com'},
                     {'rank': '700', 'domain': 'itmedia.co.jp'},
                     {'rank': '701', 'domain': 'ebay.fr'},
                     {'rank': '702', 'domain': 'blog.me'},
                     {'rank': '703', 'domain': 'nordstrom.com'},
                     {'rank': '704', 'domain': 'olx.pl'},
                     {'rank': '705', 'domain': 'lanacion.com.ar'},
                     {'rank': '706', 'domain': 'as.com'},
                     {'rank': '707', 'domain': 'ouedkniss.com'},
                     {'rank': '708', 'domain': 'badoo.com'},
                     {'rank': '709', 'domain': 'ria.ru'},
                     {'rank': '710', 'domain': 'infobae.com'},
                     {'rank': '711', 'domain': 'cambridge.org'},
                     {'rank': '712', 'domain': 'rbc.ru'},
                     {'rank': '713', 'domain': 'google.com.do'},
                     {'rank': '714', 'domain': 'makemytrip.com'},
                     {'rank': '715', 'domain': 'google.rs'},
                     {'rank': '716', 'domain': 'realclearpolitics.com'},
                     {'rank': '717', 'domain': 'getbootstrap.com'},
                     {'rank': '718', 'domain': 'nikkei.com'},
                     {'rank': '719', 'domain': 'genius.com'},
                     {'rank': '720', 'domain': 'grammarly.com'},
                     {'rank': '721', 'domain': 'imagebam.com'},
                     {'rank': '722', 'domain': 'shaparak.ir'},
                     {'rank': '723', 'domain': 'siteadvisor.com'},
                     {'rank': '724', 'domain': 'cbsnews.com'},
                     {'rank': '725', 'domain': 'subito.it'},
                     {'rank': '726', 'domain': 'suning.com'},
                     {'rank': '727', 'domain': 'voc.com.cn'},
                     {'rank': '728', 'domain': 'redirectvoluum.com'},
                     {'rank': '729', 'domain': 'india.com'},
                     {'rank': '730', 'domain': 'ca.gov'},
                     {'rank': '731', 'domain': 'drive2.ru'},
                     {'rank': '732', 'domain': 'vporn.com'},
                     {'rank': '733', 'domain': 'libero.it'},
                     {'rank': '734', 'domain': 'albawabhnews.com'},
                     {'rank': '735', 'domain': 'utorrent.com'},
                     {'rank': '736', 'domain': 'paytm.in'},
                     {'rank': '737', 'domain': 'the-watch-series.to'},
                     {'rank': '738', 'domain': 'www.gov.uk'},
                     {'rank': '739', 'domain': 'appledaily.com.tw'},
                     {'rank': '740', 'domain': 'investopedia.com'},
                     {'rank': '741', 'domain': 'clarin.com'},
                     {'rank': '742', 'domain': 'cnbc.com'},
                     {'rank': '743', 'domain': 'ruliweb.com'},
                     {'rank': '744', 'domain': 'donga.com'},
                     {'rank': '745', 'domain': 'biobiochile.cl'},
                     {'rank': '746', 'domain': 'eyny.com'},
                     {'rank': '747', 'domain': 'sabq.org'},
                     {'rank': '748', 'domain': 'dribbble.com'},
                     {'rank': '749', 'domain': 'bola.net'},
                     {'rank': '750', 'domain': 'discogs.com'},
                     {'rank': '751', 'domain': 'thesportbible.com'},
                     {'rank': '752', 'domain': 'google.com.my'},
                     {'rank': '753', 'domain': 'nypost.com'},
                     {'rank': '754', 'domain': 'quikr.com'},
                     {'rank': '755', 'domain': 'streamcloud.eu'},
                     {'rank': '756', 'domain': 'europa.eu'},
                     {'rank': '757', 'domain': 'emol.com'},
                     {'rank': '758', 'domain': 'youjizz.com'},
                     {'rank': '759', 'domain': 'vnexpress.net'},
                     {'rank': '760', 'domain': 'momoshop.com.tw'},
                     {'rank': '761', 'domain': 'khanacademy.org'},
                     {'rank': '762', 'domain': 'patch.com'},
                     {'rank': '763', 'domain': 'drtuber.com'},
                     {'rank': '764', 'domain': 'merriam-webster.com'},
                     {'rank': '765', 'domain': 'okcupid.com'},
                     {'rank': '766', 'domain': 'youboy.com'},
                     {'rank': '767', 'domain': 'gazeta.pl'},
                     {'rank': '768', 'domain': 'gotporn.com'},
                     {'rank': '769', 'domain': 'torrentkim3.net'},
                     {'rank': '770', 'domain': 'bodybuilding.com'},
                     {'rank': '771', 'domain': 'prothom-alo.com'},
                     {'rank': '772', 'domain': 'thekitchn.com'},
                     {'rank': '773', 'domain': 'chron.com'},
                     {'rank': '774', 'domain': 'elbalad.news'},
                     {'rank': '775', 'domain': 'ultimate-guitar.com'},
                     {'rank': '776', 'domain': 'thehindu.com'},
                     {'rank': '777', 'domain': 'slickdeals.net'},
                     {'rank': '778', 'domain': 'norton.com'},
                     {'rank': '779', 'domain': 'google.by'},
                     {'rank': '780', 'domain': 'gstatic.com'},
                     {'rank': '781', 'domain': 'discordapp.com'},
                     {'rank': '782', 'domain': 'oschina.net'},
                     {'rank': '783', 'domain': 'olx.ua'},
                     {'rank': '784', 'domain': 'google.com.ec'},
                     {'rank': '785', 'domain': 'likes.com'},
                     {'rank': '786', 'domain': 'wish.com'},
                     {'rank': '787', 'domain': 'linkshrink.net'},
                     {'rank': '788', 'domain': 'codecanyon.net'},
                     {'rank': '789', 'domain': 'ccm.net'},
                     {'rank': '790', 'domain': 'watchseries.ac'},
                     {'rank': '791', 'domain': 'pof.com'},
                     {'rank': '792', 'domain': 'ameba.jp'},
                     {'rank': '793', 'domain': 'nudevista.com'},
                     {'rank': '794', 'domain': 'abcnews.go.com'},
                     {'rank': '795', 'domain': 'cdiscount.com'},
                     {'rank': '796', 'domain': 'leadzupc.com'},
                     {'rank': '797', 'domain': 'custhelp.com'},
                     {'rank': '798', 'domain': 'theodysseyonline.com'},
                     {'rank': '799', 'domain': 'apache.org'},
                     {'rank': '800', 'domain': 'pixiv.net'},
                     {'rank': '801', 'domain': 'topix.com'},
                     {'rank': '802', 'domain': 'bhphotovideo.com'},
                     {'rank': '803', 'domain': 'lapatilla.com'},
                     {'rank': '804', 'domain': 'stanford.edu'},
                     {'rank': '805', 'domain': 'trulia.com'},
                     {'rank': '806', 'domain': '1337x.to'},
                     {'rank': '807', 'domain': 'gigazine.net'},
                     {'rank': '808', 'domain': 'nhk.or.jp'},
                     {'rank': '809', 'domain': 'manoramaonline.com'},
                     {'rank': '810', 'domain': 'jeuxvideo.com'},
                     {'rank': '811', 'domain': 'istockphoto.com'},
                     {'rank': '812', 'domain': 'folha.uol.com.br'},
                     {'rank': '813', 'domain': 'allrecipes.com'},
                     {'rank': '814', 'domain': 'ted.com'},
                     {'rank': '815', 'domain': 'porn.com'},
                     {'rank': '816', 'domain': 'popmyads.com'},
                     {'rank': '817', 'domain': 'inven.co.kr'},
                     {'rank': '818', 'domain': 'elevenia.co.id'},
                     {'rank': '819', 'domain': 'cnmo.com'},
                     {'rank': '820', 'domain': 'realtor.com'},
                     {'rank': '821', 'domain': 'cookpad.com'},
                     {'rank': '822', 'domain': 'informer.com'},
                     {'rank': '823', 'domain': 'intel.com'},
                     {'rank': '824', 'domain': 'turbobit.net'},
                     {'rank': '825', 'domain': 'le.com'},
                     {'rank': '826', 'domain': 'ensonhaber.com'},
                     {'rank': '827', 'domain': 'entrepreneur.com'},
                     {'rank': '828', 'domain': 'deviantart.net'},
                     {'rank': '829', 'domain': 'adme.ru'},
                     {'rank': '830', 'domain': 'flaticon.com'},
                     {'rank': '831', 'domain': 'allocine.fr'},
                     {'rank': '832', 'domain': 'flirt4free.com'},
                     {'rank': '833', 'domain': 'sputniknews.com'},
                     {'rank': '834', 'domain': 'bidverdrs.com'},
                     {'rank': '835', 'domain': 'bluehost.com'},
                     {'rank': '836', 'domain': 'imagetwist.com'},
                     {'rank': '837', 'domain': 'bleacherreport.com'},
                     {'rank': '838', 'domain': 'gazzetta.it'},
                     {'rank': '839', 'domain': 'fmovies.to'},
                     {'rank': '840', 'domain': 'k618.cn'},
                     {'rank': '841', 'domain': 'convert2mp3.net'},
                     {'rank': '842', 'domain': 'nanoadexchange.com'},
                     {'rank': '843', 'domain': 'fanfiction.net'},
                     {'rank': '844', 'domain': 'pcmag.com'},
                     {'rank': '845', 'domain': 'tmz.com'},
                     {'rank': '846', 'domain': 'zara.com'},
                     {'rank': '847', 'domain': 'shareasale.com'},
                     {'rank': '848', 'domain': 'ablogica.com'},
                     {'rank': '849', 'domain': 'gmarket.co.kr'},
                     {'rank': '850', 'domain': 'bandcamp.com'},
                     {'rank': '851', 'domain': 'mynavi.jp'},
                     {'rank': '852', 'domain': 'trackingclick.net'},
                     {'rank': '853', 'domain': 'kayak.com'},
                     {'rank': '854', 'domain': 'thevideo.me'},
                     {'rank': '855', 'domain': 'lowes.com'},
                     {'rank': '856', 'domain': 'constantcontact.com'},
                     {'rank': '857', 'domain': 'googlevideo.com'},
                     {'rank': '858', 'domain': 'techradar.com'},
                     {'rank': '859', 'domain': 'jquery.com'},
                     {'rank': '860', 'domain': 'wired.com'},
                     {'rank': '861', 'domain': 'gmanetwork.com'},
                     {'rank': '862', 'domain': 'harvard.edu'},
                     {'rank': '863', 'domain': 'gamespot.com'},
                     {'rank': '864', 'domain': 'zhanqi.tv'},
                     {'rank': '865', 'domain': 'fbsbx.com'},
                     {'rank': '866', 'domain': 'stockstar.com'},
                     {'rank': '867', 'domain': '58.com'},
                     {'rank': '868', 'domain': 'adp.com'},
                     {'rank': '869', 'domain': 'aweber.com'},
                     {'rank': '870', 'domain': 'jsfiddle.net'},
                     {'rank': '871', 'domain': 'blogimg.jp'},
                     {'rank': '872', 'domain': 'cqnews.net'},
                     {'rank': '873', 'domain': 'liveleak.com'},
                     {'rank': '874', 'domain': 'google.hr'},
                     {'rank': '875', 'domain': 'mangafox.me'},
                     {'rank': '876', 'domain': 'gazetaexpress.com'},
                     {'rank': '877', 'domain': 'ptt.cc'},
                     {'rank': '878', 'domain': 'altervista.org'},
                     {'rank': '879', 'domain': 'google.tn'},
                     {'rank': '880', 'domain': 'r7mediar.com'},
                     {'rank': '881', 'domain': 'digitaltrends.com'},
                     {'rank': '882', 'domain': 'fidelity.com'},
                     {'rank': '883', 'domain': '360doc.com'},
                     {'rank': '884', 'domain': 'clicksor.net'},
                     {'rank': '885', 'domain': 'onlinevideoconverter.com'},
                     {'rank': '886', 'domain': 'dict.cc'},
                     {'rank': '887', 'domain': 'kohls.com'},
                     {'rank': '888', 'domain': 'absoluteclickscom.com'},
                     {'rank': '889', 'domain': 'blogspot.de'},
                     {'rank': '890', 'domain': 'microsoftstore.com'},
                     {'rank': '891', 'domain': 'discuss.com.hk'},
                     {'rank': '892', 'domain': 'slimspots.com'},
                     {'rank': '893', 'domain': 'oneindia.com'},
                     {'rank': '894', 'domain': 'southwest.com'},
                     {'rank': '895', 'domain': 'asahi.com'},
                     {'rank': '896', 'domain': 'agoda.com'},
                     {'rank': '897', 'domain': 'namesakeoscilloscopemarquis.com'},
                     {'rank': '898', 'domain': 'epwk.com'},
                     {'rank': '899', 'domain': 'buzzfil.net'},
                     {'rank': '900', 'domain': 'duolingo.com'},
                     {'rank': '901', 'domain': 'chouftv.ma'},
                     {'rank': '902', 'domain': 'zytpirwai.net'},
                     {'rank': '903', 'domain': 'criteo.com'},
                     {'rank': '904', 'domain': 'indiamart.com'},
                     {'rank': '905', 'domain': 'unity3d.com'},
                     {'rank': '906', 'domain': 'envato.com'},
                     {'rank': '907', 'domain': 'okta.com'},
                     {'rank': '908', 'domain': 'so-net.ne.jp'},
                     {'rank': '909', 'domain': 'backpage.com'},
                     {'rank': '910', 'domain': 'americanas.com.br'},
                     {'rank': '911', 'domain': 'nairaland.com'},
                     {'rank': '912', 'domain': 'thatviralfeed.net'},
                     {'rank': '913', 'domain': 'xtube.com'},
                     {'rank': '914', 'domain': 'narod.ru'},
                     {'rank': '915', 'domain': 'slate.com'},
                     {'rank': '916', 'domain': 'vodlocker.com'},
                     {'rank': '917', 'domain': 'fontawesome.io'},
                     {'rank': '918', 'domain': 'fishki.net'},
                     {'rank': '919', 'domain': 'twoo.com'},
                     {'rank': '920', 'domain': 'yandex.com.tr'},
                     {'rank': '921', 'domain': 'ancestry.com'},
                     {'rank': '922', 'domain': 'google.co.ao'},
                     {'rank': '923', 'domain': 'nexusmods.com'},
                     {'rank': '924', 'domain': 'g2a.com'},
                     {'rank': '925', 'domain': 'people.com'},
                     {'rank': '926', 'domain': 'yaplakal.com'},
                     {'rank': '927', 'domain': 'codeonclick.com'},
                     {'rank': '928', 'domain': 'nydailynews.com'},
                     {'rank': '929', 'domain': 'idnes.cz'},
                     {'rank': '930', 'domain': 'hh.ru'},
                     {'rank': '931', 'domain': 'junbi-tracker.com'},
                     {'rank': '932', 'domain': 'dl-protect.com'},
                     {'rank': '933', 'domain': 'ps4ux.com'},
                     {'rank': '934', 'domain': 'aiomp3.com'},
                     {'rank': '935', 'domain': 'java.com'},
                     {'rank': '936', 'domain': 'keepvid.com'},
                     {'rank': '937', 'domain': 'sapo.pt'},
                     {'rank': '938', 'domain': 'verizon.com'},
                     {'rank': '939', 'domain': 'tempo.co'},
                     {'rank': '940', 'domain': 'costco.com'},
                     {'rank': '941', 'domain': '11st.co.kr'},
                     {'rank': '942', 'domain': 'liveinternet.ru'},
                     {'rank': '943', 'domain': 'ticketmaster.com'},
                     {'rank': '944', 'domain': 'gutefrage.net'},
                     {'rank': '945', 'domain': 'wattpad.com'},
                     {'rank': '946', 'domain': 'scribol.com'},
                     {'rank': '947', 'domain': 'zopim.com'},
                     {'rank': '948', 'domain': 'thisav.com'},
                     {'rank': '949', 'domain': 'mobile01.com'},
                     {'rank': '950', 'domain': 'caixa.gov.br'},
                     {'rank': '951', 'domain': 'zedo.com'},
                     {'rank': '952', 'domain': 'udn.com'},
                     {'rank': '953', 'domain': 'ssl-images-amazon.com'},
                     {'rank': '954', 'domain': 'axisbank.co.in'},
                     {'rank': '955', 'domain': 'motherless.com'},
                     {'rank': '956', 'domain': 'pixlr.com'},
                     {'rank': '957', 'domain': 'op.gg'},
                     {'rank': '958', 'domain': 'houzz.com'},
                     {'rank': '959', 'domain': 'list.tmall.com'},
                     {'rank': '960', 'domain': 'rutube.ru'},
                     {'rank': '961', 'domain': 'himado.in'},
                     {'rank': '962', 'domain': 'javedch.com'},
                     {'rank': '963', 'domain': 'xunlei.com'},
                     {'rank': '964', 'domain': 'gap.com'},
                     {'rank': '965', 'domain': 'safefinder.com'},
                     {'rank': '966', 'domain': 'mgid.com'},
                     {'rank': '967', 'domain': 'marketwatch.com'},
                     {'rank': '968', 'domain': 'google.com.ly'},
                     {'rank': '969', 'domain': 'gotowebinar.com'},
                     {'rank': '970', 'domain': 'getmyads.com'},
                     {'rank': '971', 'domain': 'lequipe.fr'},
                     {'rank': '972', 'domain': 'givemesport.com'},
                     {'rank': '973', 'domain': 'cam4.com'},
                     {'rank': '974', 'domain': 'lolesports.com'},
                     {'rank': '975', 'domain': 'voyeurhit.com'},
                     {'rank': '976', 'domain': 'socialblade.com'},
                     {'rank': '977', 'domain': 'ebay.es'},
                     {'rank': '978', 'domain': 'angel.co'},
                     {'rank': '979', 'domain': 'detonate.com'},
                     {'rank': '980', 'domain': 'truecaller.com'},
                     {'rank': '981', 'domain': '52pk.com'},
                     {'rank': '982', 'domain': 'fanpage.gr'},
                     {'rank': '983', 'domain': 'hepsiburada.com'},
                     {'rank': '984', 'domain': 'eonline.com'},
                     {'rank': '985', 'domain': 'pocofh.com'},
                     {'rank': '986', 'domain': 'nature.com'},
                     {'rank': '987', 'domain': 'tomshardware.com'},
                     {'rank': '988', 'domain': 'jb51.net'},
                     {'rank': '989', 'domain': 'theatlantic.com'},
                     {'rank': '990', 'domain': 'hostgator.com'},
                     {'rank': '991', 'domain': 'xbox.com'},
                     {'rank': '992', 'domain': 'tnaflix.com'},
                     {'rank': '993', 'domain': 'uptobox.com'},
                     {'rank': '994', 'domain': 'klikbca.com'},
                     {'rank': '995', 'domain': 'biblegateway.com'},
                     {'rank': '996', 'domain': 'reallifecam.com'},
                     {'rank': '997', 'domain': 'retailmenot.com'},
                     {'rank': '998', 'domain': 'ibtimes.co.uk'},
                     {'rank': '999', 'domain': 'myfreecams.com'},
                     {'rank': '1000', 'domain': 'ilbe.com'}]

    job_feeder(pkg_resources.resource_filename("pathspider", "tests/data/dnsresolvtest.csv"), spider)
    assert spider.was_shutdown
    print (spider.jobs)
    print (expected_jobs)
    assert spider.jobs == expected_jobs




from io import BytesIO

import pycurl

from pathspider.base import CONN_OK
from pathspider.base import CONN_FAILED

def connect_http(source, job, conn_timeout, curlopts=None, curlinfos=None):
    """
    This helper function will perform a TCP connection. It will not perform
    any special action in the event that this is the experimental flow,
    but can be customised on a per-call basis through the curlopts argument.
    """

    if 'dp' not in job:
        job['dp'] = 80

    c = pycurl.Curl()

    if curlopts is None:
        curlopts = {}

    if source is not None:
        if ":" in job['dip']:
            curlopts[pycurl.INTERFACE] = source[1]
        else:
            curlopts[pycurl.INTERFACE] = source[0]

    if ':' in job['dip']:
        ipString = '[' + job['dip'] + ']'
    else:
        ipString = job['dip']

    if pycurl.URL not in curlopts:
        if 'domain' in job:
            curlopts[pycurl.URL] = "http://" + job['domain'] + ":" + str(job['dp']) + "/"
        else:
            curlopts[pycurl.URL] = "http://" + ipString + ":" + str(job['dp']) + "/"

    if pycurl.USERAGENT not in curlopts:
        useragent = "PATHspider (https://pathspider.net/)"
        curlopts[pycurl.USERAGENT] = useragent

    curlopts[pycurl.TIMEOUT] = conn_timeout

    curlopts[pycurl.CONNECT_TO] = ["::{}:{}".format(ipString, job['dp'])]

    header = BytesIO()
    curlopts[pycurl.HEADERFUNCTION] = header.write
    body = BytesIO()
    curlopts[pycurl.WRITEDATA] = body

    curlopts[pycurl.FRESH_CONNECT] = 1
    curlopts[pycurl.FORBID_REUSE] = 1

    for o in curlopts:
        try:
            c.setopt(o, curlopts[o])
        except TypeError:
            return {'spdr_state': CONN_FAILED, 'sp': 0}

    try:
        c.perform()
        sp = c.getinfo(pycurl.LOCAL_PORT)
        code = c.getinfo(pycurl.RESPONSE_CODE)
        info = {}
        if curlinfos is not None:
            for curlinfo in curlinfos:
                info[curlinfo] = c.getinfo(curlinfo)
        c.close()
        return {
            'sp': sp,
            'spdr_state': CONN_OK,
            'http_response_code': code,
            'http_response_header': header.getvalue().decode('utf-8', 'backslashreplace'),
            'http_response_body': body.getvalue().decode('utf-8', 'backslashreplace'),
            'http_info': info,
        }
    except pycurl.error: # TODO: Catch timeout seperately
        return {'spdr_state': CONN_FAILED, 'sp': 0}

def connect_https(source, job, conn_timeout, curlopts=None, curlinfos=None):
    if 'dp' not in job:
        job['dp'] = 443

    if curlopts is None:
        curlopts = {}

    if ':' in job['dip']:
        ipString = '[' + job['dip'] + ']'
    else:
        ipString = job['dip']

    if pycurl.URL not in curlopts:
        if 'domain' in job:
            url = "http://" + job['domain'] + ":" + str(job['dp']) + "/"
        else:
            url = "http://" + ipString + ":" + str(job['dp']) + "/"
    else:
        curlopts[pycurl.URL] = url

    if pycurl.SSL_VERIFYHOST not in curlopts:
        curlopts[pycurl.SSL_VERIFYHOST] = 0

    if pycurl.SSL_VERIFYPEER not in curlopts:
        curlopts[pycurl.SSL_VERIFYPEER] = 0

    return connect_http(source, job, conn_timeout, curlopts, curlinfos)




from pathspider.tests.chains import ChainTestCase

class TestObserverMatching(ChainTestCase):

    def test_observer_random_flowcount(self):
        test_trace = "random.pcap"
        self.create_observer(test_trace, [])
        flows = self.run_observer()
        assert len(flows) == 0
    
    def test_observer_real_flowcount(self):
        test_trace = "real.pcap"
        self.create_observer(test_trace, [])
        flows = self.run_observer()
        assert len(flows) == 6317
    
    def test_observer_icmp_ttl_flowcount(self):
        test_trace = "icmp_ttl.pcap"
        self.create_observer(test_trace, [])
        flows = self.run_observer()
        assert len(flows) == 297




from collections import namedtuple
from tempfile import NamedTemporaryFile

from scapy.all import wrpcap

from pathspider.chains.dscp import DSCPChain
from pathspider.plugins.dscp import DSCP
from pathspider.tests.chains import ChainTestCase
from pathspider.base import CONN_OK
from pathspider.base import CONN_FAILED

def test_plugin_dscp_combine_not_observed():
    for valid in [True, False]:
        flows = [
                 {'observed': True},
                 {'observed': False}
                ]
        conditions = DSCP.combine_flows(None, flows)
        assert "pathspider.not_observed" in conditions

def test_plugin_dscp_combine():
    test_groups = [
                   (CONN_OK,  CONN_OK,  "dscp.46.connectivity.works"),
                   (CONN_OK,  CONN_FAILED, "dscp.46.connectivity.broken"),
                   (CONN_FAILED, CONN_OK,  "dscp.46.connectivity.transient"),
                   (CONN_FAILED, CONN_FAILED, "dscp.46.connectivity.offline")
                  ]
    for group in test_groups:
        flows = [
                 {'observed': True, 'spdr_state': group[0], 'dscp_mark_syn_fwd': 0, 'dscp_mark_data_fwd': 0, 'dscp_mark_syn_rev': 0, 'dscp_mark_data_rev': 0},
                 {'observed': True, 'spdr_state': group[1], 'dscp_mark_syn_fwd': 46, 'dscp_mark_data_fwd': 46, 'dscp_mark_syn_rev': 46, 'dscp_mark_data_rev': 46}
                ]
        conditions = DSCP.combine_flows(None, flows)
        assert group[2] in conditions
        if 'dscp.46.connectivity.works' in conditions:
            assert 'dscp.0.replymark:0' in conditions
            assert 'dscp.46.replymark:46' in conditions






from pathspider.tests.chains import ChainTestCase

from pathspider.chains.mss import MSSChain

class TestMSSChain(ChainTestCase):

    def test_chain_mss_ipv4(self):
        test_trace = "mss_ipv4.pcap"
        self.create_observer(test_trace, [MSSChain])

        expected_mss = {
            'mss_len_fwd': 4,
            'mss_len_rev': 4,
            'mss_value_fwd': 1460,
            'mss_value_rev': 1452
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_mss:
            assert flows[0][key] == expected_mss[key]

    def test_chain_mss_ipv6(self):
        test_trace = "mss_ipv6.pcap"
        self.create_observer(test_trace, [MSSChain])

        expected_mss = {
            'mss_len_fwd': 4,
            'mss_len_rev': 4,
            'mss_value_fwd': 1440,
            'mss_value_rev': 1360
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_mss:
            assert flows[0][key] == expected_mss[key]

    def test_chain_mss_none(self):
        test_trace = "mss_none.pcap"
        self.create_observer(test_trace, [MSSChain])

        expected_mss = {
            'mss_len_fwd': None,
            'mss_len_rev': None,
            'mss_value_fwd': None,
            'mss_value_rev': None
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_mss:
            assert flows[0][key] == expected_mss[key]

    def test_chain_mss_3bytes(self):
        test_trace = "mss_3bytes.pcap"
        self.create_observer(test_trace, [MSSChain])

        expected_mss = {
            'mss_len_fwd': 5,
            'mss_len_rev': None,
            'mss_value_fwd': 1460,
            'mss_value_rev': None
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_mss:
            assert flows[0][key] == expected_mss[key]



"""
.. module:: pathspider.chains.tcp
   :synopsis: A flow analysis chain for ICMP messages and useful ICMP related
              constants

This module contains the ICMPChain flow analysis chain which can be used by
PATHspider's Observer for recording ICMPv4 [RFC792]_ and ICMPv6 [RFC4443]_
details.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain

#: ICMPv4 Message Type - Unreachable
ICMP4_UNREACHABLE = 3
#: ICMPv4 Message Type - TTL Exceeded
ICMP4_TTLEXCEEDED = 11

#: ICMPv6 Message Type - Unreachable
ICMP6_UNREACHABLE = 1
#: ICMPv6 Message Type - Time Exceeded
ICMP6_TTLEXCEEDED = 3

class ICMPChain(Chain):
    """
    This flow analysis chain records details of ICMP messages in
    the flow record. It will record when a message of certain types have been
    seen during a flow.

    +----------------------+--------+-------------------------------------------------------------+
    | Field Name           | Type   | Meaning                                                     |
    +======================+========+=============================================================+
    | ``icmp_unreachable`` | bool   | An ICMP unreachable message was seen in the reverse         |
    |                      |        | direction                                                   |
    +----------------------+--------+-------------------------------------------------------------+
    """

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``False``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['icmp_unreachable'] = False
        return True

    def icmp4(self, rec, ip, q, rev): # pylint: disable=no-self-use,unused-argument
        """
        Records ICMPv4 details.

        ICMPv4 Unreachable Messages
            Sets ``icmp_unreachable`` to ``True`` if an ICMP Unreachable
            message is seen in the reverse direction.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv4 packet that was observed to be part of this flow
                   and contained an ICMPv4 header
        :type ip: plt.ip
        :param q: the ICMP quotation of the packet that triggered this message
                  (if any)
        :type q: plt.ip
        :param rev: ``True`` if the packet was in the reverse direction,
                    ``False`` if in the forward direction
        :type rev: bool
        :return: ``False`` if an ICMP unreachable message has been observed,
                 otherwise ``True``
        :rtype: bool
        """

        if rev and ip.icmp.type == ICMP4_UNREACHABLE:
            rec['icmp_unreachable'] = True
        return not rec['icmp_unreachable']

    def icmp6(self, rec, ip6, q, rev): # pylint: disable=no-self-use,unused-argument
        """
        Records ICMPv6 details.

        ICMPv6 Unreachable Messages
            Sets ``icmp_unreachable`` to ``True`` if an ICMP Unreachable
            message is seen in the reverse direction.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv6 packet that was observed to be part of this flow
                   and contained an ICMPv6 header
        :type ip: plt.ip6
        :param q: the ICMP quotation of the packet that triggered this message
                  (if any)
        :type q: plt.ip
        :param rev: ``True`` if the packet was in the reverse direction,
                    ``False`` if in the forward direction
        :type rev: bool
        :return: ``False`` if an ICMP unreachable message has been observed,
                 otherwise ``True``
        :rtype: bool
        """

        if rev and ip6.icmp6.type == ICMP6_UNREACHABLE:
            rec['icmp_unreachable'] = True
        return not rec['icmp_unreachable']




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.icmp import ICMPChain

class TestICMPChain(ChainTestCase):

    def test_chain_icmp_single_unreachable(self):
        # Note that this will register as in the FORWARD direction
        test_trace = "icmp_single_unreachable.pcap"
        self.create_observer(test_trace, [ICMPChain])

        expected_icmp = {
            'icmp_unreachable': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1
    
        for key in expected_icmp:
            assert flows[0][key] == expected_icmp[key]


    def test_chain_icmp_single_unreachable_nopayload(self):
        # Note that this will register as in the FORWARD direction
        test_trace = "icmp_single_unreachable_nopayload.pcap"
        self.create_observer(test_trace, [ICMPChain])

        expected_icmp = {
            'icmp_unreachable': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1
    
        for key in expected_icmp:
            assert flows[0][key] == expected_icmp[key]

    def test_chain_icmp_ipv6_unreachable(self):
        test_trace = "icmp_ipv6_unreachable.pcap"
        self.create_observer(test_trace, [ICMPChain])

        expected_icmp = {
            'icmp_unreachable': True,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_icmp:
            assert flows[0][key] == expected_icmp[key]

    def test_chain_icmp_ttl_exceeded(self):
        test_trace = "icmp_ttl_exceeded.pcap"
        self.create_observer(test_trace, [ICMPChain])

        expected_icmp = {
            'icmp_unreachable': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_icmp:
            assert flows[0][key] == expected_icmp[key]

    def test_chain_icmp_synack_unreachable(self):
        test_trace = "icmp_synack_unreachable.pcap"
        self.create_observer(test_trace, [ICMPChain])

        expected_icmp = {
            'icmp_unreachable': True,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_icmp:
            assert flows[0][key] == expected_icmp[key]



import stem
import stem.control
import pycurl

from pathspider.base import CONN_DISCARD
from pathspider.helpers.http import connect_http


def connect_tor_http(controller,
                     circuit_path,
                     job,
                     conn_timeout,
                     curlopts=None,
                     curlinfos=None):
    """
    This helper function will perform an HTTP request over Tor. It will not
    perform any special action in the event that this is the experimental flow,
    but can be customised on a per-call basis through the curlopts argument.
    """

    if curlopts is None:
        curlopts = {}

    curlopts[pycurl.PROXY] = "localhost"
    curlopts[pycurl.PROXYPORT] = 9050
    curlopts[pycurl.PROXYTYPE] = pycurl.PROXYTYPE_SOCKS5_HOSTNAME

    attach_error = []

    try:
        if circuit_path is not None:
            circuit_path = circuit_path.split(",")
        circuit_id = controller.new_circuit(circuit_path, await_build=True)
    except stem.CircuitExtensionFailed:
        return {"spdr_state": CONN_DISCARD}

    def attach_stream(stream):
        try:
            if stream.status == 'NEW':
                if (stream.target_address == job['dip'] and
                        stream.target_port == job['dp']):
                    controller.attach_stream(stream.id, circuit_id)
        except stem.OperationFailed:
            attach_error.append(None)

    controller.add_event_listener(attach_stream, stem.control.EventType.STREAM) # pylint: disable=no-member
    result = connect_http(None, job, conn_timeout, curlopts, curlinfos)
    controller.remove_event_listener(attach_stream)

    if len(attach_error) > 0:
        return {"spdr_state": CONN_DISCARD}

    return result




from io import BytesIO
import json

import pycurl
from pyroute2 import IPDB # pylint: disable=no-name-in-module

def interface_up(ifname):
    with IPDB() as ipdb:
        for i in set(ipdb.interfaces.values()):
            if i.ifname == ifname and i.operstate == 'UP':
                return True
        return False

def ipv4_address(ifname):
    with IPDB() as ipdb:
        addrset = [x[0] for x in ipdb.interfaces[ifname].ipaddr if '.' in x[0]]
        if len(addrset) > 0:
            # Should return the first IPv4 address of the interface...if there are any
            return addrset[0]

def ipv6_address(ifname):
    with IPDB() as ipdb:
        addrset = [x[0] for x in ipdb.interfaces[ifname].ipaddr
                   if ':' in x[0] and not x[0].startswith('fe')]
        if len(addrset) > 0:
            # Should return the first IPv6 address of the interface...if there are any
            return addrset[0]

def ipv4_address_public(ifname):
    c = pycurl.Curl()
    body = BytesIO()
    c.setopt(c.URL, "https://stat.ripe.net/data/whats-my-ip/data.json")
    c.setopt(c.INTERFACE, ifname)
    c.setopt(c.WRITEDATA, body)
    c.setopt(c.IPRESOLVE, c.IPRESOLVE_V4)
    c.perform()
    return json.loads((body.getvalue()).decode('utf-8'))['data']['ip']

def ipv6_address_public(ifname):
    try:
        c = pycurl.Curl()
        body = BytesIO()
        c.setopt(c.URL, "https://stat.ripe.net/data/whats-my-ip/data.json")
        c.setopt(c.INTERFACE, ifname)
        c.setopt(c.WRITEDATA, body)
        c.setopt(c.IPRESOLVE, c.IPRESOLVE_V6)
        c.perform()
        return json.loads((body.getvalue()).decode('utf-8'))['data']['ip']
    except pycurl.error:
        return "::"

def ipv4_asn(ifname):
    c = pycurl.Curl()
    body = BytesIO()
    c.setopt(c.URL, "https://stat.ripe.net/data/prefix-overview/data.json?resource={}"
             .format(ipv4_address_public(ifname)))
    c.setopt(c.INTERFACE, ifname)
    c.setopt(c.WRITEDATA, body)
    c.perform()
    asns = json.loads((body.getvalue()).decode('utf-8'))['data']['asns']
    if len(asns) == 1:
        return asns[0]['asn']
    else:
        return None

def ipv6_asn(ifname):
    try:
        c = pycurl.Curl()
        body = BytesIO()
        c.setopt(c.URL, "https://stat.ripe.net/data/prefix-overview/data.json?resource={}"
                 .format(ipv6_address_public(ifname)))
        c.setopt(c.INTERFACE, ifname)
        c.setopt(c.WRITEDATA, body)
        c.perform()
        asns = json.loads((body.getvalue()).decode('utf-8'))['data']['asns']
        if len(asns) == 1:
            return asns[0]['asn']
        else:
            return None
    except pycurl.error:
        return None



import socket

import pathspider.base
from pathspider.base import PluggableSpider
from pathspider.base import CONN_DISCARD
from pathspider.desync import DesynchronizedSpider


class DNSResolv(DesynchronizedSpider, PluggableSpider):

    name = "dnsresolv"
    description = "Simple Input List DNS Resolver"
    version = pathspider.base.__version__
    chains = [] # Use the dummy observer

    def resolv_host(self, job, config): # pylint: disable=unused-argument
        if 'domain' not in job or job['domain'] is None:
            return {'spdr_state': CONN_DISCARD} # Always discard
        try:
            ips = set([str(i[4][0]) for i in socket.getaddrinfo(job['domain'], 80)])
        except socket.gaierror:
            return {'spdr_state': CONN_DISCARD} # Always discard
        except OSError:
            return {'spdr_state': CONN_DISCARD} # Always discard
        for ip in ips:
            res = job.copy()
            res['dip'] = ip
            self.outqueue.put(res)
        return {'spdr_state': CONN_DISCARD} # Always discard

    connections = [resolv_host]



"""
.. module:: pathspider.chains.ecn
   :synopsis: A flow analysis chain for Explicit Congestion Notification

This module contains the ECNChain flow analysis chain which can be used by
PATHspider's Observer for recording Explicit Congestion Notification [RFC3168]_
details.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain
from pathspider.chains.tcp import TCP_SYN

class ECNChain(Chain):
    """
    This flow analysis chain records details of ECN markings in the IP header.

    +-------------------+------+----------------------------------------------------+
    | Field Name        | Type | Meaning                                            |
    +===================+======+====================================================+
    | ecn_ect0_syn_fwd  | bool | An ECT0 mark was seen in the forward direction     |
    |                   |      | on a TCP SYN packet                                |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect1_syn_fwd  | bool | An ECT1 mark was seen in the forward direction     |
    |                   |      | on a TCP SYN packet                                |
    +-------------------+------+----------------------------------------------------+
    | ecn_ce_syn_fwd    | bool | An CE mark was seen in the forward direction       |
    |                   |      | on a TCP SYN packet                                |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect0_data_fwd | bool | An ECT0 mark was seen in the forward direction     |
    |                   |      | on a TCP packet with a payload or a non-TCP packet |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect1_data_fwd | bool | An ECT1 mark was seen in the forward direction     |
    |                   |      | on a TCP packet with a payload or a non-TCP packet |
    +-------------------+------+----------------------------------------------------+
    | ecn_ce_data_fwd   | bool | An CE mark was seen in the forward direction       |
    |                   |      | on a TCP packet with a payload or a non-TCP packet |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect0_syn_rev  | bool | An ECT0 mark was seen in the reverse direction     |
    |                   |      | on a TCP SYN packet                                |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect1_syn_rev  | bool | An ECT1 mark was seen in the reverse direction     |
    |                   |      | on a TCP SYN packet                                |
    +-------------------+------+----------------------------------------------------+
    | ecn_ce_syn_rev    | bool | An CE mark was seen in the reverse direction       |
    |                   |      | on a TCP SYN packet                                |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect0_data_rev | bool | An ECT0 mark was seen in the reverse direction     |
    |                   |      | on a TCP packet with a payload or a non-TCP packet |
    +-------------------+------+----------------------------------------------------+
    | ecn_ect1_data_rev | bool | An ECT1 mark was seen in the reverse direction     |
    |                   |      | on a TCP packet with a payload or a non-TCP packet |
    +-------------------+------+----------------------------------------------------+
    | ecn_ce_data_rev   | bool | An CE mark was seen in the reverse direction       |
    |                   |      | on a TCP packet with a payload or a non-TCP packet |
    +-------------------+------+----------------------------------------------------+
    """

    def new_flow(self, rec, ip): # pylint: disable=unused-argument
        """
        For a new flow, all fields will be initialised to ``False``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        for d in ['fwd', 'rev']:
            for t in ['syn', 'data']:
                for f in ['ect0', 'ect1', 'ce']:
                    rec['ecn_{}_{}_{}'.format(f, t, d)] = False

        return True

    def ip4(self, rec, ip, rev):
        """
        Records ECN markings from an IPv4 header.

        ECN Marking
            If an ECT0, ECT1 or CE mark is seen in the IPv4 header, the relevant
            field will be set to ``True``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv4 packet that was observed to be part of this flow
        :type ip: plt.ip
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always ``True``
        :rtype: bool
        """

        return self._ecn_extract(rec, ip, rev)

    def ip6(self, rec, ip, rev):
        """
        Records ECN markings from an IPv6 header.

        ECN Marking
            If an ECT0, ECT1 or CE mark is seen in the IPv6 header, the relevant
            field will be set to ``True``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv6 packet that was observed to be part of this flow
        :type ip: plt.ip6
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always ``True``
        :rtype: bool
        """

        return self._ecn_extract(rec, ip, rev)

    def _ecn_extract(self, rec, ip, rev):
        ECT_ZERO = 0x02
        ECT_ONE = 0x01
        ECT_CE = 0x03

        ipmark = None

        if ip.traffic_class & ECT_CE == ECT_ZERO:
            ipmark = 'ecn_ect0'
        if ip.traffic_class & ECT_CE == ECT_ONE:
            ipmark = 'ecn_ect1'
        if ip.traffic_class & ECT_CE == ECT_CE:
            ipmark = 'ecn_ce'

        if ipmark is not None:
            if ip.tcp and ip.tcp.flags & TCP_SYN == TCP_SYN:
                t = 'syn'
            else:
                t = 'data'
            d = 'rev' if rev else 'fwd'
            rec['{}_{}_{}'.format(ipmark, t, d)] = True

        return True



"""
.. module:: pathspider.chains.basic
   :synopsis: A flow analysis chain for basic TCP/IP flow information

This module contains the BasicChain flow analysis chain which can be used by
PATHspider's Observer for recording source and destination addresses and
packet/octet counts.

"""

from pathspider.chains.base import Chain

class BasicChain(Chain):
    """
    This flow analysis chain records details from the TCP/IP headers.

    +------------+------+-----------------------------------------------------+
    | Field Name | Type | Meaning                                             |
    +============+======+=====================================================+
    | dip        | str  | Layer 3 (IPv4/IPv6) source address                  |
    +------------+------+-----------------------------------------------------+
    | sp         | int  | Layer 4 (TCP/UDP) source port                       |
    +------------+------+-----------------------------------------------------+
    | dp         | int  | Layer 4 (TCP/UDP) destination port                  |
    +------------+------+-----------------------------------------------------+
    | pkt_fwd    | int  | A count of the number of packets seen in the        |
    |            |      | forward direction                                   |
    +------------+------+-----------------------------------------------------+
    | pkt_rev    | int  | A count of the number of packets seen in the        |
    |            |      | reverse direction                                   |
    +------------+------+-----------------------------------------------------+
    | oct_fwd    | int  | A count of the number of octets seen in the         |
    |            |      | forward direction                                   |
    +------------+------+-----------------------------------------------------+
    | oct_rev    | int  | A count of the number of octets seen in the         |
    |            |      | reverse direction                                   |
    +------------+------+-----------------------------------------------------+
    """

    def _extract_ports(self, ip):
        if ip.udp:
            return (ip.udp.src_port, ip.udp.dst_port)
        elif ip.tcp:
            return (ip.tcp.src_port, ip.tcp.dst_port)
        else:
            return (None, None)

    def new_flow(self, rec, ip):
        """
        New flow function that sets up basic flow information
        """

        # Extract addresses and ports
        (rec['sip'], rec['dip'], rec['proto']) = (str(ip.src_prefix), str(ip.dst_prefix), ip.proto)
        (rec['sp'], rec['dp']) = self._extract_ports(ip)

        # Initialize counters
        rec['pkt_fwd'] = 0
        rec['pkt_rev'] = 0
        rec['oct_fwd'] = 0
        rec['oct_rev'] = 0

        # we want to keep this flow
        return True

    def ip4(self, rec, ip, rev):
        return self._basic_count(rec, ip, rev)

    def ip6(self, rec, ip, rev):
        return self._basic_count(rec, ip, rev)

    def _basic_count(self, rec, ip, rev):
        """
        Packet function that counts packets and octets per flow
        """

        if rev:
            rec["pkt_rev"] += 1
            rec["oct_rev"] += ip.size
        else:
            rec["pkt_fwd"] += 1
            rec["oct_fwd"] += ip.size

        return True



import time
import logging
import queue
import uuid

from pathspider.base import Spider
from pathspider.base import CONN_DISCARD
from pathspider.base import QUEUE_SLEEP
from pathspider.base import SHUTDOWN_SENTINEL


class DesynchronizedSpider(Spider):
    # pylint: disable=W0223

    connections = []

    def __init__(self, worker_count, libtrace_uri, args, server_mode=False):
        super().__init__(worker_count, libtrace_uri, args, server_mode)

        self.__logger = logging.getLogger('desync')

        self._config_count = len(self.connections)

        if hasattr(self.args, 'connect') and self.args.connect.startswith('tor'):
            self.__logger.warning("Clamping worker count to one to avoid overloading Tor!")
            self.worker_count = 1
            self.__logger.warning("Observer is not much use with Tor, disabling!")
            self.chains = []

    def configurator(self):
        """
        Since there is no need for a configurator thread in a
        desynchronized spider, this thread is a no-op
        """
        self.__logger.info("configurations are not synchronized")

    def worker(self, worker_number):
        """
        This function provides the logic for
        configuration-synchronized worker threads.

        :param worker_number: The unique number of the worker.
        :type worker_number: int

        The workers operate as continuous loops:

         * Fetch next job from the job queue
         * Perform pre-connection operations
         * Acquire a lock for "config_zero"
         * Perform the "config_zero" connection
         * Release "config_zero"
         * Acquire a lock for "config_one"
         * Perform the "config_one" connection
         * Release "config_one"
         * Perform post-connection operations for config_zero and pass the
           result to the merger
         * Perform post-connection operations for config_one and pass the
           result to the merger
         * Do it all again

        If the job fetched is the SHUTDOWN_SENTINEL, then the worker will
        terminate as this indicates that all the jobs have now been processed.
        """

        worker_active = True

        while self.running:
            if worker_active:
                try:
                    job = self.jobqueue.get_nowait()
                    jobId = uuid.uuid1().hex

                    # Break on shutdown sentinel
                    if job == SHUTDOWN_SENTINEL:
                        self.jobqueue.task_done()
                        self.__logger.debug(
                            "shutting down worker %d on sentinel",
                            worker_number)
                        worker_active = False
                        with self.active_worker_lock:
                            self.active_worker_count -= 1
                            self.__logger.debug("%d workers still active",
                                                self.active_worker_count)
                        continue

                    self.__logger.debug("got a job: " + repr(job))
                except queue.Empty:
                    time.sleep(QUEUE_SLEEP)
                else:
                    # Hook for preconnection
                    self.pre_connect(job)

                    conns = []
                    should_discard = False

                    for config in range(0, len(self.connections)):
                        conn = self._connect_wrapper(
                            job, config, connect=self.connections[config])
                        if 'spdr_state' in conn:
                            if conn['spdr_state'] == CONN_DISCARD:
                                should_discard = True
                        conns.append(conn)

                    if not should_discard:
                        # Save job record for combiner
                        self.jobtab[jobId] = job

                        # Pass results on for merge
                        self._finalise_conns(job, jobId, conns)

                    self.__logger.debug("job complete: " + repr(job))
                    self.jobqueue.task_done()
            elif not self.stopping:
                time.sleep(QUEUE_SLEEP)
            else:
                break

    @classmethod
    def register_args(cls, subparsers):
        # pylint: disable=no-member
        parser = subparsers.add_parser(cls.name, help=cls.description)
        parser.add_argument(
            "--timeout",
            default=5,
            type=int,
            help=("The timeout to use for attempted "
                  "connections in seconds (Default: 5)"))
        if hasattr(cls, "connect_supported"):
            parser.add_argument(
                "--connect",
                type=str,
                choices=cls.connect_supported,
                default=cls.connect_supported[0],
                metavar="[{}]".format("|".join(cls.connect_supported)),
                help="Type of connection to perform (Default: {})".format(
                    cls.connect_supported[0]))
            for connect in cls.connect_supported:
                if connect.startswith('tor'):
                    parser.add_argument(
                        "--tor-path",
                        type=str,
                        help="A comma-seperated list of Tor relay fingerprints to use for building circuits"
                    )
                    break
        parser.set_defaults(spider=cls)
        if hasattr(cls, "extra_args"):
            cls.extra_args(parser)




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.evil import EvilChain

class TestEvilChain(ChainTestCase):

    def test_chain_evilbit_tcp_none(self):
        test_trace = "evilbit_tcp_none.pcap"
        self.create_observer(test_trace, [EvilChain])

        expected_evil = {
            'evilbit_syn_fwd': False,
            'evilbit_syn_rev': False,
            'evilbit_data_fwd': None,
            'evilbit_data_rev': None,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_evil:
            assert flows[0][key] == expected_evil[key]

    def test_chain_evilbit_udp_none(self):
        test_trace = "evilbit_udp_none.pcap"
        self.create_observer(test_trace, [EvilChain])

        expected_evil = {
            'evilbit_syn_fwd': None,
            'evilbit_syn_rev': None,
            'evilbit_data_fwd': False,
            'evilbit_data_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_evil:
            assert flows[0][key] == expected_evil[key]


    def test_chain_evilbit_tcp_fwd(self):
        test_trace = "evilbit_tcp_fwd.pcap"
        self.create_observer(test_trace, [EvilChain])

        expected_evil = {
            'evilbit_syn_fwd': True,
            'evilbit_syn_rev': False,
            'evilbit_data_fwd': None,
            'evilbit_data_rev': None,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_evil:
            assert flows[0][key] == expected_evil[key]

    def test_chain_evilbit_tcp_rev(self):
        test_trace = "evilbit_tcp_rev.pcap"
        self.create_observer(test_trace, [EvilChain])

        expected_evil = {
            'evilbit_syn_fwd': False,
            'evilbit_syn_rev': True,
            'evilbit_data_fwd': None,
            'evilbit_data_rev': None,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_evil:
            assert flows[0][key] == expected_evil[key]


    def test_chain_evilbit_udp_fwd(self):
        test_trace = "evilbit_udp_fwd.pcap"
        self.create_observer(test_trace, [EvilChain])

        expected_evil = {
            'evilbit_syn_fwd': None,
            'evilbit_syn_rev': None,
            'evilbit_data_fwd': True,
            'evilbit_data_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_evil:
            assert flows[0][key] == expected_evil[key]

    def test_chain_evilbit_udp_rev(self):
        test_trace = "evilbit_udp_rev.pcap"
        self.create_observer(test_trace, [EvilChain])

        expected_evil = {
            'evilbit_syn_fwd': None,
            'evilbit_syn_rev': None,
            'evilbit_data_fwd': False,
            'evilbit_data_rev': True,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_evil:
            assert flows[0][key] == expected_evil[key]





import logging
import subprocess

import pathspider.base
from pathspider.base import CONN_OK
from pathspider.base import PluggableSpider
from pathspider.sync import SynchronizedSpider
from pathspider.chains.basic import BasicChain
from pathspider.chains.tcp import TCPChain
from pathspider.chains.dscp import DSCPChain
from pathspider.chains.dns import DNSChain

class DSCP(SynchronizedSpider, PluggableSpider):

    name = "dscp"
    description = "Differentiated Services Codepoints"
    version = pathspider.base.__version__
    chains = [BasicChain, DSCPChain, TCPChain, DNSChain]
    connect_supported = ["http", "tcp", "dnstcp", "dnsudp"]

    def config_no_dscp(self):  # pylint: disable=no-self-use
        """
        Disables DSCP marking via iptables.
        """

        logger = logging.getLogger('dscp')
        for iptables in ['iptables', 'ip6tables']:
            subprocess.check_call([iptables, '-t', 'mangle', '-F'])
        logger.debug("Configurator disabled DSCP marking")

    def config_dscp(self):
        """
        Enables DSCP marking via iptables.
        """
        logger = logging.getLogger('dscp')
        for iptables in ['iptables', 'ip6tables']:
            subprocess.check_call([
                iptables, '-t', 'mangle', '-A', 'OUTPUT', '-j', 'DSCP',
                '--set-dscp', str(self.args.codepoint)
            ])
        logger.debug("Configurator enabled DSCP marking")

    configurations = [config_no_dscp, config_dscp]

    def combine_flows(self, flows):
        # discard non-observed flows
        for f in flows:
            if not f['observed']:
                return ['pathspider.not_observed']

        conditions = []

        baseline = 'dscp.' + str(flows[0]['dscp_mark_syn_fwd'] or
                                 flows[0]['dscp_mark_data_fwd']) + '.'
        test = 'dscp.' + str(flows[1]['dscp_mark_syn_fwd'] or
                             flows[1]['dscp_mark_data_fwd']) + '.'

        if flows[0]['spdr_state'] == CONN_OK and flows[1][
                'spdr_state'] == CONN_OK:
            cond_conn = test + 'connectivity.works'
        elif flows[0]['spdr_state'] == CONN_OK and not flows[1][
                'spdr_state'] == CONN_OK:
            cond_conn = test + 'connectivity.broken'
        elif not flows[0]['spdr_state'] == CONN_OK and flows[1][
                'spdr_state'] == CONN_OK:
            cond_conn = test + 'connectivity.transient'
        else:
            cond_conn = test + 'connectivity.offline'
        conditions.append(cond_conn)

        baseline_replymark = flows[0]['dscp_mark_syn_rev'] or flows[0]['dscp_mark_data_rev']
        test_replymark = flows[1]['dscp_mark_syn_rev'] or flows[1]['dscp_mark_data_rev']

        if baseline_replymark is not None:
            conditions.append(baseline + 'replymark:' + str(baseline_replymark))
        if test_replymark is not None:
            conditions.append(test + 'replymark:' + str(test_replymark))

        return conditions

    @staticmethod
    def extra_args(parser):
        parser.add_argument(
            "--codepoint",
            type=int,
            choices=range(0, 64),
            default='48',
            metavar="[0-63]",
            help="DSCP codepoint to send (Default: 48)")




import logging

from scapy.all import send

from pathspider.desync import DesynchronizedSpider

from pathspider.chains.basic import BasicChain

class ForgeSpider(DesynchronizedSpider):

    chains = [BasicChain]
    packets = 0

    def __init__(self, worker_count, libtrace_uri, args):
        super().__init__(worker_count, libtrace_uri, args)

        self.__logger = logging.getLogger('forge')
        self._config_count = self.packets
        self.connections = [self.connect] * self.packets # pylint: disable=no-member

    def pre_connect(self, job):
        self.setup(job)

    def setup(self, job):
        pass

    def connect(self, job, seq):
        pkt = self.forge(job, seq)
        send(pkt, verbose=0)
        return {'sp': pkt.getlayer(1).sport}

    def forge(self, job, config):
        raise NotImplementedError("Cannot register an abstract plugin")

    @classmethod
    def register_args(cls, subparsers):
        # pylint: disable=no-member
        parser = subparsers.add_parser(cls.name, help=cls.description)
        parser.set_defaults(spider=cls)
        if hasattr(cls, "connect_supported"):
            parser.add_argument(
                "--connect",
                type=str,
                choices=cls.connect_supported,
                default=cls.connect_supported[0],
                metavar="[{}]".format("|".join(cls.connect_supported)),
                help="Type of connection to perform (Default: {})".format(
                    cls.connect_supported[0]))

        if hasattr(cls, "extra_args"):
            cls.extra_args(parser)



"""
.. module:: pathspider.chains.tcp
   :synopsis: A flow observer chain for basic TCP behaviour, TCP options
              parser and useful TCP related constants

This module contains the TCPChain flow analysis chain which can be used by
PATHspider's Observer for recording basic TCP [RFC793]_ behaviour details. This
module also contains a helper function that may be used by chains for the
parsing of TCP options and a number of useful TCP related constants that can be
used to interpret the results added to flow records by TCPChain.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>
.. codeauthor:: Piet De Vaere <piet@devae.re>

"""

from pathspider.chains.base import Chain

#: TCP Flag - CWR
TCP_CWR = 0x80
#: TCP Flag - ECE
TCP_ECE = 0x40
#: TCP Flag - URG
TCP_URG = 0x20
#: TCP Flag - ACK
TCP_ACK = 0x10
#: TCP Flag - PSH
TCP_PSH = 0x08
#: TCP Flag - RST
TCP_RST = 0x04
#: TCP Flag - SYN
TCP_SYN = 0x02
#: TCP Flag - FIN
TCP_FIN = 0x01

#: TCP Flags - SYN and ACK
TCP_SA = (TCP_SYN | TCP_ACK)
#: TCP Flags - SYN, ACK and ECE
TCP_SEC = (TCP_SYN | TCP_ECE | TCP_CWR)
#: TCP Flags - SYN, ACK, ECE and CWR
TCP_SAEC = (TCP_SYN | TCP_ACK | TCP_ECE | TCP_CWR)
#: TCP Flags - SYN, ACK, ECE
TCP_SAE = (TCP_SYN | TCP_ACK | TCP_ECE)

#: TCP Option - End of options list
TO_EOL = 0
#: TCP Option - No Operation
TO_NOP = 1
#: TCP Option - Maximum Segment Size
TO_MSS = 2
#: TCP Option - Window Scaling
TO_WS = 3
#: TCP Option - Selective Acknowledgement Permitted
TO_SACKOK = 4
#: TCP Option - Selective Acknowledgement
TO_SACK = 5
#: TCP Option - Timestamp
TO_TS = 8
#: TCP Option - Multipath TCP
TO_MPTCP = 30
#: TCP Option - TCP Fast Open Cookie
TO_FASTOPEN = 34
#: TCP Option - Experimental Option A
TO_EXPA = 254
#: TCP Option - Experimental Option B
TO_EXPB = 255
#: TCP Option Experiment ID - TCP Fast Open
TO_EXID_FASTOPEN = (0xF9, 0x89)

def tcp_options(tcp):
    """
    Parses and extracts TCP options from a python-libtrace TCP object.

    .. warning:: This is a pure Python implementation of a TCP options parser
                 and does not benefit from the speed advantage generally
                 realised by calling to libtrace functions written in C through
                 python-libtrace.

    :param tcp: The TCP header to extract options from
    :type tcp: plt.tcp
    :returns: A mapping of option kinds to values
    :rtype: dict
    """

    optbytes = tcp.data[20:tcp.doff*4]
    opthash = {}

    # parse options in place
    cp = 0
    ncp = 0

    while cp < len(optbytes):
        # skip NOP
        if optbytes[cp] == TO_NOP:
            cp += 1
            continue
        # die on EOL
        if optbytes[cp] == TO_EOL:
            break

        # parse options length
        ncp = cp + optbytes[cp+1]

        # copy options data into hash
        # FIXME doesn't handle multiples (#189)
        opthash[optbytes[cp]] = optbytes[cp+2:ncp]

        # advance
        cp = ncp

    return opthash

class TCPChain(Chain):
    """
    This flow analysis chain records details of basic TCP behaviour in the
    flow record. It will determine when a 3WHS has completed and has simplified
    logic for determining when a TCP flow has completed.

    +----------------------+------+---------------------------------------------------------+
    | Field Name           | Type | Description                                             |
    +======================+======+=========================================================+
    | ``tcp_synflags_fwd`` | int  | SYN flags seen in the forward direction                 |
    +----------------------+------+---------------------------------------------------------+
    | ``tcp_synflags_rev`` | int  | SYN flags seen in the reverse direction                 |
    +----------------------+------+---------------------------------------------------------+
    | ``tcp_fin_fwd``      | bool | At least one FIN flag was seen in the forward direction |
    +----------------------+------+---------------------------------------------------------+
    | ``tcp_fin_rev``      | bool | At least one FIN flag was seen in the reverse direction |
    +----------------------+------+---------------------------------------------------------+
    | ``tcp_rst_fwd``      | bool | At least one RST flag was seen in the forward direction |
    +----------------------+------+---------------------------------------------------------+
    | ``tcp_rst_rev``      | bool | At least one RST flag was seen in the reverse direction |
    +----------------------+------+---------------------------------------------------------+
    | ``tcp_connected``    | bool | The 3WHS completed                                      |
    +----------------------+------+---------------------------------------------------------+
    """

    def new_flow(self, rec, ip): # pylint: disable=W0613
        """
        For a new flow, all fields will be initialised to ``False`` except
        ``tcp_synflags_*`` which will be set to ``None``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['tcp_synflags_fwd'] = None
        rec['tcp_synflags_rev'] = None

        rec['tcp_fin_fwd'] = False
        rec['tcp_fin_rev'] = False
        rec['tcp_rst_fwd'] = False
        rec['tcp_rst_rev'] = False

        rec['tcp_connected'] = False

        return True

    def tcp(self, rec, tcp, rev):
        """
        Records basic TCP behaviour details.

        SYN Flags
            This will record the SYN flags observed in each direction. These will
            not be recorded again if there are futher segments in the flow with a
            SYN bit set, the first SYN observed wins.

        FIN and RST Flags
            If a segment has the FIN or RST flags, the relevant fields are set
            to true.

        3WHS
            If a SYN was observed in the forward direction, and a SYNACK in the
            reverse direction and the segment passed is an ACK in the forward
            direction then ``tcp_connected`` will be set to True.

        Flow Completion
            If a FIN has been observed in one direction and this segment
            contains a FIN in the other direction, a flow is considered
            complete. If a RST has been observed in either direction, a flow is
            considered complete.

        :param rec: the flow record
        :type rec: dict
        :param tcp: the TCP segment that was observed to be part of this flow
        :type ip: plt.tcp
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        if tcp.syn_flag:
            rec['tcp_synflags_rev' if rev else 'tcp_synflags_fwd'] = tcp.flags

        # This test is intended to catch the completion of the 3WHS.
        if (not rec['tcp_connected'] and rev == 0 and
                rec['tcp_synflags_fwd'] is not None and
                rec['tcp_synflags_rev'] is not None and
                rec['tcp_synflags_fwd'] & TCP_SYN == TCP_SYN and
                rec['tcp_synflags_rev'] & TCP_SA == TCP_SA and
                tcp.ack_flag):
            rec['tcp_connected'] = True

        if tcp.fin_flag and rev:
            rec['tcp_fin_fwd'] = True
        if tcp.fin_flag and not rev:
            rec['tcp_fin_rev'] = True
        if tcp.rst_flag and rev:
            rec['tcp_rst_rev'] = True
        if tcp.rst_flag and not rev:
            rec['tcp_rst_fwd'] = True

        return not ((rec['tcp_fin_fwd'] and rec['tcp_fin_rev']) or
                    rec['tcp_rst_fwd'] or rec['tcp_rst_rev'])



import collections
import logging
import base64
import queue
import math

from pathspider.base import SHUTDOWN_SENTINEL


def _flow4_ids(ip):
    # FIXME keep map of fragment IDs to keys (#144)

    icmp_with_payload = {3, 4, 5, 11, 12}
    quotation_fid = False
    if ip.proto == 1 and ip.icmp.type in icmp_with_payload:
        if ip.icmp.payload is not None and len(ip.icmp.payload.data) >= 20:
            ip = ip.icmp.payload
            quotation_fid = True

    protos_with_ports = {6, 17, 132, 136}
    if ip.proto in protos_with_ports:
        # key includes ports
        fid = ip.src_prefix.addr + ip.dst_prefix.addr + ip.data[
            9:10] + ip.payload[0:4]
        rid = ip.dst_prefix.addr + ip.src_prefix.addr + ip.data[
            9:10] + ip.payload[2:4] + ip.payload[0:2]
    else:
        # no ports, just 3-tuple
        fid = ip.src_prefix.addr + ip.dst_prefix.addr + ip.data[9:10]
        rid = ip.dst_prefix.addr + ip.src_prefix.addr + ip.data[9:10]

    if quotation_fid:
        # If the fid is based on an ICMP quotation, need to be reversed
        return (base64.b64encode(rid), base64.b64encode(fid))
    else:
        return (base64.b64encode(fid), base64.b64encode(rid))


def _flow6_ids(ip6):
    icmp_with_payload = {1, 2, 3, 4}
    quotation_fid = False

    if ip6.proto == 58 and ip6.icmp6.type in icmp_with_payload:
        if ip6.icmp6.payload is not None and len(ip6.icmp6.payload.data) >= 40:
            ip6 = ip6.icmp6.payload
            quotation_fid = True

    protos_with_ports = {6, 17, 132, 136}
    if ip6.proto in protos_with_ports:
        # key includes ports
        fid = ip6.src_prefix.addr + ip6.dst_prefix.addr + ip6.data[
            6:7] + ip6.payload[0:4]
        rid = ip6.dst_prefix.addr + ip6.src_prefix.addr + ip6.data[
            6:7] + ip6.payload[2:4] + ip6.payload[0:2]
    else:
        # no ports, just 3-tuple
        fid = ip6.src_prefix.addr + ip6.dst_prefix.addr + ip6.data[6:7]
        rid = ip6.dst_prefix.addr + ip6.src_prefix.addr + ip6.data[6:7]

    if quotation_fid:
        # If the fid is based on an ICMP quotation, need to be reversed
        return (base64.b64encode(rid), base64.b64encode(fid))
    else:
        return (base64.b64encode(fid), base64.b64encode(rid))


PacketClockTimer = collections.namedtuple("PacketClockTimer", ("time", "fn"))


class Observer:
    """
    Wraps a packet source identified by a libtrace URI,
    parses packets to divide them into flows, passing these
    packets and flows onto a function chain to allow
    data to be associated with each flow.
    """

    def __init__(self, lturi, chains=None, idle_timeout=30, expiry_timeout=5):
        """
        Create an Observer.

        :param chains: Array of Observer chain classes
        :see also: :ref:`Observer Documentation <observer>`
        """

        # Only import this when needed
        import plt as libtrace

        # Control
        self._irq = None
        self._irq_fired = False

        # Libtrace initialization
        self._trace = libtrace.trace(lturi)  # pylint: disable=no-member
        self._trace.start()
        self._pkt = libtrace.packet()  # pylint: disable=no-member

        # Chains of functions to evaluate
        chains = chains if chains is not None else []
        self._chains = [chain() for chain in chains]

        # Packet timer and bintables
        self._ptq = 0  # current packet timer, quantized
        self._idle_bins = {}  # map bin number to set of fids
        self._expiry_bins = {}  # map bin number to set of fids
        self._idle_timeout = idle_timeout
        self._expiry_timeout = expiry_timeout
        self._bin_quantum = 1

        #self._tq = []                  # packet timer queue (heap)

        # Flow tables
        self._active = {}
        self._expiring = {}
        self._ignored = set()

        # Emitter queue
        self._emitted = collections.deque()

        # Statistics and logging
        self._logger = logging.getLogger("observer")
        self._ct_pkt = 0
        self._ct_nonip = 0
        self._ct_shortkey = 0
        self._ct_ignored = 0
        self._ct_flow = 0

    def _interrupted(self):
        if not self._irq_fired and self._irq is not None:
            try:
                self._irq.get_nowait()
                self._irq_fired = True
            except queue.Empty:
                pass

        return self._irq_fired

    def _get_chains(self, name):
        return [
            c.__getattribute__(name) for c in self._chains if hasattr(c, name)
        ]

    def _next_packet(self):
        # Import only when needed
        import plt as libtrace

        # see if someone told us to stop
        if self._interrupted():
            return False

        # see if we're done iterating
        if not self._trace.read_packet(self._pkt):
            return False

        # count the packet
        self._ct_pkt += 1

        # advance the packet clock
        self._tick(self._pkt.seconds)

        # get a flow ID and associated flow record for the packet
        (fid, rec, rev) = self._get_flow()

        # don't dispatch if we don't have a record
        # (this happens for non-IP packets and flows
        #  we know we want to ignore)
        if not rec:
            return True

        keep_flow = True

        # run IP header chains
        if self._pkt.ip:
            for fn in self._get_chains("ip4"):
                keep_flow = keep_flow and fn(rec, self._pkt.ip, rev=rev)
            if self._pkt.icmp:
                for fn in self._get_chains("icmp4"):
                    q = self._pkt.icmp.payload # pylint: disable=no-member
                    keep_flow = keep_flow and fn(rec, self._pkt.ip, q, rev=rev)

        elif self._pkt.ip6:
            for fn in self._get_chains("ip6"):
                keep_flow = keep_flow and fn(rec, self._pkt.ip6, rev=rev)
            if self._pkt.icmp6:
                for fn in self._get_chains("icmp6"):
                    q = self._pkt.icmp6.payload # pylint: disable=no-member
                    keep_flow = keep_flow and fn(
                        rec, self._pkt.ip6, q, rev=rev)

        # run transport header chains
        if self._pkt.tcp:
            for fn in self._get_chains("tcp"):
                keep_flow = keep_flow and fn(rec, self._pkt.tcp, rev=rev)
        elif self._pkt.udp:
            for fn in self._get_chains("udp"):
                keep_flow = keep_flow and fn(rec, self._pkt.udp, rev=rev)

        # complete the flow if any chain function asked us to
        if not keep_flow:
            self._flow_complete(fid)

        # we processed a packet, keep going
        return True

    # def _set_timer(self, delay, fid):
    #     # add to queue
    #     heapq.heappush(self._tq, PacketClockTimer(self._pt + delay,
    #                    self._finish_expiry_tfn(fid)))

    def _get_flow(self):
        """
        Get a flow record for the given packet.
        Create a new basic flow record
        """
        # get possible a flow IDs for the packet
        try:
            if self._pkt.ip:
                (ffid, rfid) = _flow4_ids(self._pkt.ip)
                ip = self._pkt.ip
            elif self._pkt.ip6:
                (ffid, rfid) = _flow6_ids(self._pkt.ip6)
                ip = self._pkt.ip6
            else:
                # we don't care about non-IP packets
                self._ct_nonip += 1
                return (None, None, False)
        except ValueError:
            self._ct_shortkey += 1
            return (None, None, False)

        # now look for forward and reverse in ignored, active,
        # and expiring tables.
        if ffid in self._ignored:
            return (None, None, False)
        elif rfid in self._ignored:
            return (None, None, False)
        elif ffid in self._active:
            (fid, rec, active) = (ffid, self._active[ffid], True)
            #self._logger.debug("found forward flow for "+str(ffid))
        elif ffid in self._expiring:
            (fid, rec, active) = (ffid, self._expiring[ffid], False)
            #self._logger.debug("found expiring forward flow for "+str(ffid))
        elif rfid in self._active:
            (fid, rec, active) = (rfid, self._active[rfid], True)
            #self._logger.debug("found reverse flow for "+str(rfid))
        elif rfid in self._expiring:
            (fid, rec, active) = (rfid, self._expiring[rfid], False)
            #self._logger.debug("found expiring reverse flow for "+str(rfid))
        else:
            # nowhere to be found. new flow.
            rec = {'pkt_first': ip.seconds, '_idle_bin': 0}
            for fn in self._get_chains("new_flow"):
                if not fn(rec, ip):
                    # self._logger.debug("ignoring "+str(ffid))
                    self._ignored.add(ffid)
                    self._ct_ignored += 1
                    return (None, None, False)

            # wasn't vetoed. add to active table.
            fid = ffid
            self._active[ffid] = rec
            active = True
            # self._logger.debug("new flow for "+str(ffid))
            self._ct_flow += 1

        # update time and idle bin and return record
        rec['pkt_last'] = ip.seconds

        # update idle bin if we're not expiring
        if active:
            new_idle_bin = math.ceil((rec['pkt_last'] + self._idle_timeout) /
                                     self._bin_quantum) * self._bin_quantum

            if new_idle_bin > rec["_idle_bin"]:

                if rec['_idle_bin'] in self._idle_bins:
                    self._idle_bins[rec['_idle_bin']] -= set((fid, ))
                if new_idle_bin in self._idle_bins:
                    self._idle_bins[new_idle_bin] |= set((fid, ))
                else:
                    self._idle_bins[new_idle_bin] = set((fid, ))

                rec['_idle_bin'] = new_idle_bin

        return (fid, rec, bool(fid == rfid))

    def _flow_complete(self, fid):
        """
        Mark a given flow ID as complete
        """
        # skip all of this unless the flow is still in the active table
        if fid not in self._active:
            return

        # remove flow ID from idle bin
        rec = self._active[fid]
        self._idle_bins[rec['_idle_bin']] -= set((fid, ))

        del rec['_idle_bin']

        # move record to expiring table
        self._expiring[fid] = rec
        del self._active[fid]

        # assign expiry bin
        expiry_bin = math.ceil((self._ptq + self._expiry_timeout) /
                               self._bin_quantum) * self._bin_quantum
        #self._logger.debug("Completing flow "+str(fid)+" at "+str(self._ptq)+" to expire "+str(expiry_bin)+" (in "+str(expiry_bin-self._ptq)+"s)")

        if expiry_bin in self._expiry_bins:
            self._expiry_bins[expiry_bin] |= set((fid, ))
        else:
            self._expiry_bins[expiry_bin] = set((fid, ))

    def _emit_flow(self, rec):
        self._emitted.append(rec)

    def _next_flow(self):
        while len(self._emitted) == 0:
            if not self._next_packet():
                return None

        return self._emitted.popleft()

    def _tick(self, pt):
        # quantize and skip if we're not advancing
        next_ptq = math.ceil(pt / self._bin_quantum) * self._bin_quantum
        if next_ptq <= self._ptq:
            return
        elif self._ptq == 0:
            # handle zero case
            self._ptq = next_ptq
            return

        # advance quantum
        for bint in range(self._ptq + self._bin_quantum,
                          next_ptq + self._bin_quantum, self._bin_quantum):
            self._logger.debug("tick: " + str(bint))

            # process idle
            if bint in self._idle_bins:
                if len(self._idle_bins[bint]) > 0:
                    for fid in self._idle_bins[bint].copy():
                        self._flow_complete(fid)
                del self._idle_bins[bint]

            # process expiry
            if bint in self._expiry_bins:
                if len(self._expiry_bins[bint]) > 0:
                    for fid in self._expiry_bins[bint].copy():
                        self._emit_flow(self._expiring[fid])
                        del self._expiring[fid]
                del self._expiry_bins[bint]

        self._ptq = next_ptq

    # def _tick(self, pt):
    #     # Advance packet clock
    #     self._pt = pt

    #     # fire all timers whose time has come
    #     while len(self._tq) > 0 and pt > min(self._tq, key=lambda x: x.time).time:
    #         try:
    #             heapq.heappop(self._tq).fn()
    #         except:
    #             type, value, tb = sys.exc_info()
    #             traceback.print_exc()
    #             pdb.post_mortem(tb)

    # def _finish_expiry_tfn(self, fid):
    #     """
    #     On expiry timer, emit the flow
    #     and delete it from the expiring queue
    #     """
    #     def tfn():
    #         if fid in self._expiring:
    #             self._emit_flow(self._expiring[fid])
    #             del self._expiring[fid]
    #             # self._logger.debug("emitted "+str(fid)+" on expiry")
    #     return tfn

    # def purge_idle(self, timeout=30):
    #     # TODO test this, it's probably pretty slow.
    #     for fid in self._active:
    #         if self._pt - self._active['fid']['last'] > timeout:
    #             self._flow_complete(fid)

    def flush(self):
        for fid in self._expiring:
            self._emit_flow(self._expiring[fid])
            # self._logger.debug("emitted "+str(fid)+" expiring during flush")
        self._expiring.clear()

        for fid in self._active:
            self._emit_flow(self._active[fid])
            # self._logger.debug("emitted "+str(fid)+" active during flush")
        self._active.clear()

        self._ignored.clear()

    def run_flow_enqueuer(self, flowqueue, irqueue=None):
        if irqueue:
            self._irq = irqueue
            self._irq_fired = None

        # Run main loop until last packet seen
        # then flush active flows and run again
        for _ in range(2):
            while True:
                f = self._next_flow()
                if f:
                    flowqueue.put(f)
                else:
                    self.flush()
                    break

        # log observer info on shutdown
        self._logger.info(("processed %u packets "
                           "(%u dropped, %u short, %u non-ip) "
                           "into %u flows (%u ignored)"), self._ct_pkt,
                          self._trace.pkt_drops(), self._ct_shortkey,
                          self._ct_nonip, self._ct_flow, self._ct_ignored)

        flowqueue.put(SHUTDOWN_SENTINEL)


class DummyObserver:  # pylint: disable=R0903
    """
    The dummy observer provides a class compatible with the API of the Observer
    class without actually performing any operations.  This is primarily used
    for PATHspider's test suite.
    """

    def run_flow_enqueuer(self, flowqueue, irqueue=None):  # pylint: disable=R0201
        """
        When running the flow enqueuer, no network operation is performed and
        the thread will block until given a shutdown signal. When the shutdown
        signal is received it will cascade the signal onto the flowqueue
        in the same way that a real Observer instance would.
        """

        irqueue.get()
        flowqueue.put(SHUTDOWN_SENTINEL)




import multiprocessing as mp

from pathspider.base import QUEUE_SIZE
from pathspider.base import SHUTDOWN_SENTINEL
from pathspider.observer import DummyObserver

def test_dummy_observer():
    flowqueue = mp.Queue(QUEUE_SIZE)
    observer_shutdown_queue = mp.Queue(QUEUE_SIZE)

    observer = DummyObserver()
    observer_process = mp.Process(
        args=(flowqueue,
              observer_shutdown_queue),
        target=observer.run_flow_enqueuer,
        name='observer',
        daemon=True)
    observer_process.start()

    observer_shutdown_queue.put(True)

    assert flowqueue.get(True, timeout=3) == SHUTDOWN_SENTINEL

    observer_process.join(3)

    assert not observer_process.is_alive()




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.dscp import DSCPChain

class TestDSCPChain(ChainTestCase):

    def test_chain_dscp_tcp_allzero(self):
        test_trace = "dscp_tcp_allzero.pcap"
        self.create_observer(test_trace, [DSCPChain])

        expected_dscp = {
            'dscp_mark_syn_fwd': 0,
            'dscp_mark_syn_rev': 0,
            'dscp_mark_data_fwd': 0,
            'dscp_mark_data_rev': 0,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_dscp:
            assert flows[0][key] == expected_dscp[key]

    def test_chain_dscp_tcp_fwd3(self):
        test_trace = "dscp_tcp_fwd3.pcap"
        self.create_observer(test_trace, [DSCPChain])

        expected_dscp = {
            'dscp_mark_syn_fwd': 3,
            'dscp_mark_syn_rev': 0,
            'dscp_mark_data_fwd': 3,
            'dscp_mark_data_rev': 0,
        }

        flows = self.run_observer()

        assert len(flows) == 1

        for key in expected_dscp:
            assert flows[0][key] == expected_dscp[key]

    def test_chain_dscp_ipv6_tcp_fwd3(self):
        test_trace = "dscp_ipv6_tcp_fwd3.pcap"
        self.create_observer(test_trace, [DSCPChain])

        expected_dscp = {
            'dscp_mark_syn_fwd': 3,
            'dscp_mark_syn_rev': 0,
            'dscp_mark_data_fwd': 3,
            'dscp_mark_data_rev': 0,
        }

        flows = self.run_observer()

        assert len(flows) == 1

        for key in expected_dscp:
            assert flows[0][key] == expected_dscp[key]

    def test_chain_dscp_udp_allzero(self):
        test_trace = "dscp_udp_allzero.pcap"
        self.create_observer(test_trace, [DSCPChain])

        expected_dscp = {
            'dscp_mark_syn_fwd': None,
            'dscp_mark_syn_rev': None,
            'dscp_mark_data_fwd': 0,
            'dscp_mark_data_rev': 0,
        }

        flows = self.run_observer()

        assert len(flows) == 1

        for key in expected_dscp:
            assert flows[0][key] == expected_dscp[key]

    def test_chain_dscp_udp_fwd3(self):
        test_trace = "dscp_udp_fwd3.pcap"
        self.create_observer(test_trace, [DSCPChain])

        expected_dscp = {
            'dscp_mark_syn_fwd': None,
            'dscp_mark_syn_rev': None,
            'dscp_mark_data_fwd': 3,
            'dscp_mark_data_rev': 0,
        }

        flows = self.run_observer()

        assert len(flows) == 1

        for key in expected_dscp:
            assert flows[0][key] == expected_dscp[key]



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

import pathspider.base

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

setup(
    name='pathspider',
    version=pathspider.base.__version__,
    description='A tool for measuring path transparency in the Internet',
    long_description=long_description,
    url='https://pathspider.net/',
    author='Iain Learmonth',
    author_email='irl@fsfe.org',
    license='GNU GPLv2',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        # The UI
        'Environment :: Console',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Communications',
        'Topic :: Internet',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

        # Specify the Python versions you support here
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='internet measurement ecn tfo dns web www analysis router congestion path transparency latency diffserv dscp',

    packages=find_packages(exclude=['contrib', 'doc', 'examples']),
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'pspdr=pathspider.cmd.base:handle_args_wrapper',
        ],
    },
    test_suite = 'nose.collector',
    test_requires = 'nose',
)



import logging
import queue
import threading
import time
import uuid

from pathspider.base import Spider
from pathspider.base import QUEUE_SLEEP
from pathspider.base import SHUTDOWN_SENTINEL
from pathspider.helpers.tcp import connect_tcp
from pathspider.helpers.http import connect_http
from pathspider.helpers.http import connect_https
from pathspider.helpers.dns import connect_dns_tcp
from pathspider.helpers.dns import connect_dns_udp
from pathspider.base import CONN_DISCARD

class SynchronizedSpider(Spider):
    # pylint: disable=W0223

    configurations = []

    def __init__(self, worker_count, libtrace_uri, args, server_mode=False):
        super().__init__(worker_count, libtrace_uri, args, server_mode)
        self.__logger = logging.getLogger('sync')

        self._config_count = len(self.configurations)

        self.__semaphores = []

        # create semaphores for synchronizing configurations
        for config in range(0, len(self.configurations)):
            self.__semaphores.append([])
            for i in range(0, 2):
                self.__semaphores[config].append(SemaphoreN(worker_count))
                self.__semaphores[config][i].empty()

    def configurator(self):
        """
        Thread which synchronizes on a set of semaphores and alternates
        between two system states.
        """

        while self.running:
            for config in range(0, len(self.configurations)):
                self.__logger.debug("setting config %d", config)
                self.configurations[config](self)
                self.__logger.debug("config %d active", config)
                self.__semaphores[config][0].release_n(self.worker_count)
                self.__semaphores[(config + 1) % len(self.configurations)][
                    1].acquire_n(self.worker_count)

        # In case the master exits the run loop before all workers have,
        # these tokens will allow all workers to run through again,
        # until the next check at the start of the loop
        for config in range(0, len(self.configurations)):
            self.__semaphores[config][0].release_n(self.worker_count)

    def connect(self, job, config): # pylint: disable=unused-argument
        """
        Performs the requested connection.
        """

        if self.args.connect == "tcp":
            rec = connect_tcp(self.source, job, self.args.timeout)
        elif self.args.connect == "http":
            rec = connect_http(self.source, job, self.args.timeout)
        elif self.args.connect == "https":
            rec = connect_https(self.source, job, self.args.timeout)
        elif self.args.connect == "dnstcp":
            rec = connect_dns_tcp(self.source, job, self.args.timeout)
        elif self.args.connect == "dnsudp":
            rec = connect_dns_udp(self.source, job, self.args.timeout)
        else:
            raise RuntimeError("Unknown connection type requested!")

        return rec

    def worker(self, worker_number):
        """
        This function provides the logic for
        configuration-synchronized worker threads.

        :param worker_number: The unique number of the worker.
        :type worker_number: int

        The workers operate as continuous loops:

         * Fetch next job from the job queue
         * Perform pre-connection operations
         * Acquire a lock for "config_zero"
         * Perform the "config_zero" connection
         * Release "config_zero"
         * Acquire a lock for "config_one"
         * Perform the "config_one" connection
         * Release "config_one"
         * Perform post-connection operations for config_zero and pass the
           result to the merger
         * Perform post-connection operations for config_one and pass the
           result to the merger
         * Do it all again

        If the job fetched is the SHUTDOWN_SENTINEL, then the worker will
        terminate as this indicates that all the jobs have now been processed.
        """

        worker_active = True

        while self.running:
            if worker_active:
                try:
                    job = self.jobqueue.get_nowait()
                    jobId = uuid.uuid1().hex

                    # Break on shutdown sentinel
                    if job == SHUTDOWN_SENTINEL:
                        self.jobqueue.task_done()
                        self.__logger.debug(
                            "shutting down worker %d on sentinel",
                            worker_number)
                        worker_active = False
                        with self.active_worker_lock:
                            self.active_worker_count -= 1
                            self.__logger.debug("%d workers still active",
                                                self.active_worker_count)
                        continue

                    self.__logger.debug("got a job: " + repr(job))
                except queue.Empty:
                    #logger.debug("no job available, sleeping")
                    # spin the semaphores
                    for config in range(0, len(self.configurations)):
                        self.__semaphores[config][0].acquire()
                        time.sleep(QUEUE_SLEEP)
                        self.__semaphores[(config + 1) % len(
                            self.configurations)][1].release()
                else:
                    # Hook for preconnection
                    self.pre_connect(job)

                    conns = []
                    should_discard = False

                    for config in range(0, len(self.configurations)):
                        # Wait for configuration
                        self.__semaphores[config][0].acquire()

                        # Connect in configuration
                        conn = self._connect_wrapper(job, config)
                        if 'spdr_state' in conn:
                            if conn['spdr_state'] == CONN_DISCARD:
                                should_discard = True
                        conns.append(conn)

                        # Wait for next configuration
                        self.__semaphores[(config + 1) % len(
                            self.configurations)][1].release()

                    if not should_discard:
                        # Save job record for combiner
                        self.jobtab[jobId] = job

                        # Finish connections and pass on for merging
                        self._finalise_conns(job, jobId, conns)

                    self.__logger.debug("job complete: " + repr(job))
                    self.jobqueue.task_done()
            else:  # not worker_active, spin the semaphores
                for config in range(0, len(self.configurations)):
                    self.__semaphores[config][0].acquire()
                    time.sleep(QUEUE_SLEEP)
                    if config == 0:
                        with self.active_worker_lock:
                            if self.active_worker_count <= 0:
                                return
                    self.__semaphores[(config + 1) % len(self.configurations)][
                        1].release()

    @classmethod
    def register_args(cls, subparsers):
        # pylint: disable=no-member
        parser = subparsers.add_parser(cls.name, help=cls.description)
        parser.set_defaults(spider=cls)
        parser.add_argument("--connect", type=str, choices=cls.connect_supported,
                            default=cls.connect_supported[0],
                            metavar="[{}]".format("|".join(cls.connect_supported)),
                            help="Type of connection to perform (Default: {})".format(
                                cls.connect_supported[0]))
        parser.add_argument("--timeout", default=5, type=int,
                            help=("The timeout to use for attempted connections in seconds "
                                  "(Default: 5)"))
        if hasattr(cls, "extra_args"):
            cls.extra_args(parser)


class SemaphoreN(threading.BoundedSemaphore):
    """
    An extension to the standard library's BoundedSemaphore that provides
    functions to handle n tokens at once.
    """
    def __init__(self, value):
        self._value = value
        super().__init__(self._value)
        self.empty()

    def __str__(self):
        return 'SemaphoreN with a maximum value of {}.'.format(self._value)

    def acquire_n(self, value=1, blocking=True, timeout=None):
        """
        Acquire ``value`` number of tokens at once.

        The parameters ``blocking`` and ``timeout`` have the same semantics as
        :class:`BoundedSemaphore`.

        :returns: The same value as the last call to `BoundedSemaphore`'s
                  :meth:`acquire` if :meth:`acquire` were called ``value``
                  times instead of the call to this method.
        """
        ret = None
        for _ in range(value):
            ret = self.acquire(blocking=blocking, timeout=timeout)
        return ret

    def release_n(self, value=1):
        """
        Release ``value`` number of tokens at once.

        :returns: The same value as the last call to `BoundedSemaphore`'s
                  :meth:`release` if :meth:`release` were called ``value``
                  times instead of the call to this method.
        """
        ret = None
        for _ in range(value):
            ret = self.release()
        return ret

    def empty(self):
        """
        Acquire all tokens of the semaphore.
        """
        while self.acquire(blocking=False):
            pass



import logging
import subprocess

import pathspider.base
from pathspider.base import PluggableSpider
from pathspider.base import CONN_OK
from pathspider.sync import SynchronizedSpider
from pathspider.chains.basic import BasicChain
from pathspider.chains.tcp import TCPChain
from pathspider.chains.tcp import TCP_SAE
from pathspider.chains.tcp import TCP_SAEC
from pathspider.chains.ecn import ECNChain


class ECN(SynchronizedSpider, PluggableSpider):

    name = "ecn"
    description = "Explicit Congestion Notification"
    version = pathspider.base.__version__
    chains = [BasicChain, TCPChain, ECNChain]
    connect_supported = ["http", "https", "tcp", "dnstcp"]

    def config_no_ecn(self): # pylint: disable=no-self-use
        """
        Disables ECN negotiation via sysctl.
        """

        logger = logging.getLogger('ecn')
        subprocess.check_call(
            ['/sbin/sysctl', '-w', 'net.ipv4.tcp_ecn=2'],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)
        logger.debug("Configurator disabled ECN")

    def config_ecn(self): # pylint: disable=no-self-use
        """
        Enables ECN negotiation via sysctl.
        """

        logger = logging.getLogger('ecn')
        subprocess.check_call(
            ['/sbin/sysctl', '-w', 'net.ipv4.tcp_ecn=1'],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)
        logger.debug("Configurator enabled ECN")

    configurations = [config_no_ecn, config_ecn]

    def combine_flows(self, flows):
        conditions = []

        if flows[0]['spdr_state'] == CONN_OK and flows[1][
                'spdr_state'] == CONN_OK:
            conditions.append('ecn.connectivity.works')
        elif flows[0]['spdr_state'] == CONN_OK and not flows[1][
                'spdr_state'] == CONN_OK:
            conditions.append('ecn.connectivity.broken')
        elif not flows[0]['spdr_state'] == CONN_OK and flows[1][
                'spdr_state'] == CONN_OK:
            conditions.append('ecn.connectivity.transient')
        else:
            conditions.append('ecn.connectivity.offline')

        for f in flows:
            if not f['observed']:
                conditions.append('pathspider.not_observed')
                break

        if flows[1]['observed'] and flows[1]['tcp_connected']:
            if flows[1]['tcp_synflags_rev'] & TCP_SAEC == TCP_SAE:
                conditions.append('ecn.negotiation.succeeded')
            elif flows[1]['tcp_synflags_rev'] & TCP_SAEC == TCP_SAEC:
                conditions.append('ecn.negotiation.reflected')
            else:
                conditions.append('ecn.negotiation.failed')

            conditions.append('ecn.ipmark.ect0.seen' if (
                flows[1]['ecn_ect0_syn_rev'] or flows[1]['ecn_ect0_data_rev'])
                              else 'ecn.ipmark.ect0.not_seen')
            conditions.append('ecn.ipmark.ect1.seen' if (
                flows[1]['ecn_ect1_syn_rev'] or flows[1]['ecn_ect1_data_rev'])
                              else 'ecn.ipmark.ect1.not_seen')
            conditions.append('ecn.ipmark.ce.seen' if (
                flows[1]['ecn_ce_syn_rev'] or flows[1]['ecn_ce_data_rev']) else
                              'ecn.ipmark.ce.not_seen')

        return conditions
